# pylint: disable=no-member
import contextlib
import json
import subprocess
import time
import os

import attr

from .common import Driver
from .exception import ExecutionError
from ..factory import target_factory
from ..step import step
from ..util.agentwrapper import AgentWrapper
from ..util.helper import processwrapper
from ..util.managedfile import ManagedFile
from ..util.timeout import Timeout
from ..util.netns import NetNamespace
from ..resource.common import NetworkResource


@target_factory.reg_driver
@attr.s(eq=False)
class RawNetworkInterfaceDriver(Driver):
    """RawNetworkInterface - Manage a network interface and interact with it at a low level

    Args:
        manage_interface (bool, default=True): if True this driver will
        setup/teardown the interface on activate/deactivate. Set this to False
        if you are managing the interface externally.
    """

    bindings = {
        "iface": {"NetworkInterface", "RemoteNetworkInterface", "USBNetworkInterface"},
    }
    manage_interface = attr.ib(default=True, validator=attr.validators.instance_of(bool))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._record_handle = None
        self._replay_handle = None

    def on_activate(self):
        if self.manage_interface:
            self._set_interface("up")
            self._wait_state("up")

    def on_deactivate(self):
        if self.manage_interface:
            self._set_interface("down")
            self._wait_state("down")

    def _wrap_command(self, args):
        wrapper = ["sudo", "labgrid-raw-interface"]

        if self.iface.command_prefix:
            # add ssh prefix, convert command passed via ssh (including wrapper) to single argument
            return self.iface.command_prefix + [" ".join(wrapper + args)]
        else:
            # keep wrapper and args as-is
            return wrapper + args

    @step(args=["state"])
    def _set_interface(self, state):
        """Set interface to given state."""
        cmd = ["ip", self.iface.ifname, state]
        cmd = self._wrap_command(cmd)
        subprocess.check_call(cmd)

    @Driver.check_active
    def set_interface_up(self):
        """Set bound interface up."""
        self._set_interface("up")

    @Driver.check_active
    def set_interface_down(self):
        """Set bound interface down."""
        self._set_interface("down")

    def _get_state(self):
        """Returns the bound interface's operstate."""
        if_state = self.iface.extra.get("state")
        if if_state:
            return if_state

        cmd = self.iface.command_prefix + ["cat", f"/sys/class/net/{self.iface.ifname}/operstate"]
        output = processwrapper.check_output(cmd).decode("ascii")
        if_state = output.strip()
        return if_state

    @Driver.check_active
    def get_state(self):
        """Returns the bound interface's operstate."""
        return self._get_state()

    @step(title="wait_state", args=["expected_state", "timeout"])
    def _wait_state(self, expected_state, timeout=60):
        """Wait until the expected state is reached or the timeout expires."""
        timeout = Timeout(float(timeout))

        while True:
            if self._get_state() == expected_state:
                return
            if timeout.expired:
                raise TimeoutError(
                    f"exported interface {self.iface.ifname} did not go {expected_state} within {timeout.timeout} seconds"
                )
            time.sleep(1)

    @Driver.check_active
    def wait_state(self, expected_state, timeout=60):
        """Wait until the expected state is reached or the timeout expires."""
        self._wait_state(expected_state, timeout=timeout)

    @Driver.check_active
    def get_ethtool_settings(self):
        """
        Returns settings via ethtool of the bound network interface resource.
        """
        cmd = self.iface.command_prefix + ["ethtool", "--json", self.iface.ifname]
        output = subprocess.check_output(cmd, encoding="utf-8")
        return json.loads(output)[0]

    @Driver.check_active
    @step(args=["settings"])
    def ethtool_configure(self, **settings):
        """
        Change settings on interface.

        Supported settings are described in ethtool(8) --change (use "_" instead of "-").
        """
        cmd = ["ethtool", "change", self.iface.ifname]
        cmd += [item.replace("_", "-") for pair in settings.items() for item in pair]
        cmd = self._wrap_command(cmd)
        subprocess.check_call(cmd)

    @Driver.check_active
    def get_ethtool_eee_settings(self):
        """
        Returns Energy-Efficient Ethernet settings via ethtool of the bound network interface
        resource.
        """
        cmd = self.iface.command_prefix + ["ethtool", "--show-eee", "--json", self.iface.ifname]
        output = subprocess.check_output(cmd, encoding="utf-8")
        return json.loads(output)[0]

    @Driver.check_active
    @step(args=["settings"])
    def ethtool_configure_eee(self, **settings):
        """
        Change Energy-Efficient Ethernet settings via ethtool of the bound network interface
        resource.

        Supported settings are described in ethtool(8) --set-eee (use "_" instead of "-").
        """
        cmd = ["ethtool", "set-eee", self.iface.ifname]
        cmd += [item.replace("_", "-") for pair in settings.items() for item in pair]
        cmd = self._wrap_command(cmd)
        subprocess.check_call(cmd)

    @Driver.check_active
    def get_ethtool_pause_settings(self):
        """
        Returns pause parameters via ethtool of the bound network interface resource.
        """
        cmd = self.iface.command_prefix + ["ethtool", "--json", "--show-pause", self.iface.ifname]
        output = subprocess.check_output(cmd, encoding="utf-8")
        return json.loads(output)[0]

    @Driver.check_active
    @step(args=["settings"])
    def ethtool_configure_pause(self, **settings):
        """
        Change pause parameters via ethtool of the bound network interface resource.

        Supported settings are described in ethtool(8) --pause
        """
        cmd = ["ethtool", "pause", self.iface.ifname]
        cmd += [item for pair in settings.items() for item in pair]
        cmd = self._wrap_command(cmd)
        subprocess.check_call(cmd)

    def _stop(self, proc, *, timeout=None):
        assert proc is not None

        try:
            _, err = proc.communicate(timeout=timeout)
        except subprocess.TimeoutExpired:
            proc.terminate()
            _, err = proc.communicate()
            raise

        if proc.returncode:
            raise subprocess.CalledProcessError(
                returncode=proc.returncode,
                cmd=proc.args,
                stderr=err,
            )

    @Driver.check_active
    @step(args=["filename", "count", "timeout"])
    def start_record(self, filename, *, count=None, timeout=None):
        """
        Starts tcpdump on bound network interface resource.

        Args:
            filename (str): name of a file to record to, or None to record to stdout
            count (int): optional, exit after receiving this many number of packets
            timeout (int): optional, number of seconds to capture packets before tcpdump exits
        Returns:
            Popen object of tcpdump process
        """
        assert self._record_handle is None

        cmd = ["tcpdump", self.iface.ifname]
        if count is not None:
            cmd.append(str(count))
        if timeout is not None:
            cmd.append("--timeout")
            cmd.append(str(timeout))
        cmd = self._wrap_command(cmd)
        if filename is None:
            self._record_handle = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        else:
            with open(filename, "wb") as outdata:
                self._record_handle = subprocess.Popen(cmd, stdout=outdata, stderr=subprocess.PIPE)

        # wait for capture start
        stderr = b""
        while True:
            line = self._record_handle.stderr.readline()
            if not line:  # process ended prematurely
                try:
                    self._stop(self._record_handle)
                except subprocess.CalledProcessError as e:
                    # readd consumed stderr to exception
                    e.stderr = stderr
                    raise

            if line.startswith(b"tcpdump: listening on"):
                break

            # collect and log other lines in stderr before capturing has started
            stderr += line
            self.logger.warning(line.decode().rstrip())

        return self._record_handle

    @Driver.check_active
    @step(args=["timeout"])
    def stop_record(self, *, timeout=None):
        """
        Stops previously started tcpdump on bound network interface resource.

        Args:
            timeout (int): optional, maximum number of seconds to wait for the tcpdump process to
                           terminate
        """
        try:
            self._stop(self._record_handle, timeout=timeout)
        except subprocess.TimeoutExpired:
            # If live streaming packets, there is no reason to wait for tcpdump
            # to finish, so expect a timeout if piping to stdout
            if self._record_handle.stdout is None:
                raise
        finally:
            self._record_handle = None

    @contextlib.contextmanager
    def record(self, filename, *, count=None, timeout=None):
        """
        Context manager to start/stop tcpdump on bound network interface resource.

        Either count or timeout must be specified.

        Args:
            filename (str): name of a file to record to, or None to live stream packets
            count (int): optional, exit after receiving this many number of packets
            timeout (int): optional, number of seconds to capture packets before tcpdump exits
        Returns:
            Popen object of tcpdump process. If filename is None, packets can be read from stdout
        """
        assert count or timeout

        try:
            yield self.start_record(filename, count=count, timeout=timeout)
        finally:
            self.stop_record(timeout=0 if filename is None else None)

    @Driver.check_active
    @step(args=["filename"])
    def start_replay(self, filename):
        """
        Starts tcpreplay on bound network interface resource.

        Args:
            filename (str): name of a file to replay from
        Returns:
            Popen object of tcpreplay process
        """
        assert self._replay_handle is None

        if isinstance(self.iface, NetworkResource):
            mf = ManagedFile(filename, self.iface)
            mf.sync_to_resource()
            cmd = self._wrap_command([f"tcpreplay {self.iface.ifname} < {mf.get_remote_path()}"])
            self._replay_handle = subprocess.Popen(cmd, stderr=subprocess.PIPE)
        else:
            cmd = self._wrap_command(["tcpreplay", self.iface.ifname])
            with open(filename, "rb") as indata:
                self._replay_handle = subprocess.Popen(cmd, stdin=indata)

        return self._replay_handle

    @Driver.check_active
    @step(args=["timeout"])
    def stop_replay(self, *, timeout=None):
        """
        Stops previously started tcpreplay on bound network interface resource.

        Args:
            timeout (int): optional, maximum number of seconds to wait for the tcpreplay process to
                           terminate
        """
        try:
            self._stop(self._replay_handle, timeout=timeout)
        finally:
            self._replay_handle = None

    @contextlib.contextmanager
    def replay(self, filename, *, timeout=None):
        """
        Context manager to start/stop tcpreplay on bound network interface resource.

        Args:
            filename (str): name of a file to replay from
            timeout (int): optional, maximum number of seconds to wait for the tcpreplay process to
                           terminate
        """
        try:
            yield self.start_replay(filename)
        finally:
            self.stop_replay(timeout=timeout)

    @Driver.check_active
    @step()
    def get_statistics(self):
        """
        Returns basic interface statistics of bound network interface resource.
        """
        cmd = self.iface.command_prefix + ["ip", "--json", "-stats", "-stats", "link", "show", self.iface.ifname]
        output = processwrapper.check_output(cmd)
        return json.loads(output)[0]

    @Driver.check_active
    def get_address(self):
        """
        Returns the MAC address of the bound network interface resource.
        """
        return self.get_statistics()["address"]

    @Driver.check_active
    @contextlib.contextmanager
    def _netns(self, host):
        with AgentWrapper(host) as wrapper:
            ns = wrapper.load("netns")
            ns.unshare()
            yield ns

    @Driver.check_active
    @step()
    @contextlib.contextmanager
    def setup_netns(self, mac_address=None):
        with contextlib.ExitStack() as ctx:
            remote_ns = ctx.enter_context(self._netns(self.iface.host))
            local_ns = ctx.enter_context(self._netns(None))

            cmd = [
                "-d",
                "ns-macvtap",
                self.iface.ifname,
                str(remote_ns.get_pid()),
            ]
            if mac_address:
                cmd.append("--mac-address")
                cmd.append(mac_address)

            # Start tap forward in remote namespace
            remote_fwd = ctx.enter_context(
                subprocess.Popen(
                    self._wrap_command(cmd),
                    stdout=subprocess.PIPE,
                    stdin=subprocess.PIPE,
                )
            )
            ctx.callback(lambda: remote_fwd.terminate())
            self.logger.debug("Started remote tap forward '%s'", " ".join(remote_fwd.args))

            r_macaddr = None
            to = Timeout(30.0)
            while True:
                if to.expired:
                    raise TimeoutError("Timeout waiting for remote macvtap to be established")

                links = remote_ns.get_links()
                for link in links:
                    if link["ifname"] == "macvtap0":
                        r_macaddr = link["address"]
                        break

                if r_macaddr is not None:
                    break

                if remote_fwd.poll() is not None:
                    raise ExecutionError(
                        f"Remote tap forward {remote_fwd.pid} died with {remote_fwd.returncode} during setup"
                    )

                time.sleep(0.1)

            self.logger.debug("Remote macvtap MAC address is %s", r_macaddr)

            _, fd = local_ns.create_tun(address=r_macaddr)
            tun_fd = ctx.enter_context(os.fdopen(fd))

            links = local_ns.get_links()
            link_names = [link["ifname"] for link in links]
            assert "tap0" in link_names

            local_fwd = ctx.enter_context(
                subprocess.Popen(
                    local_ns.get_prefix() + ["labgrid-tap-fwd", str(tun_fd.fileno())],
                    stdin=remote_fwd.stdout,
                    stdout=remote_fwd.stdin,
                    pass_fds=(tun_fd.fileno(),),
                )
            )
            ctx.callback(lambda: local_fwd.terminate())
            self.logger.debug("Started local tap forward '%s'", " ".join(local_fwd.args))

            # Close local pipes for the remote forward, now that the local forward is running
            remote_fwd.stdin.close()
            remote_fwd.stdout.close()

            ns = NetNamespace(local_ns)

            # Wait for IPv6 address negotiation to finish
            to = Timeout(30.0)
            while True:
                if to.expired:
                    raise TimeoutError("Timeout waiting for IPv6 address negotiation to finish")

                data = json.loads(
                    ns.run(["ip", "-j", "addr", "show", "dev", ns.intf], check=True, stdout=subprocess.PIPE).stdout
                )
                for addr in data[0].get("addr_info", []):
                    if addr.get("tentative", False):
                        break
                else:
                    # No tentative addresses
                    break

                if remote_fwd.poll() is not None:
                    raise ExecutionError(f"Remote tap forward {remote_fwd.pid} died with {remote_fwd.returncode}")

                if local_fwd.poll() is not None:
                    raise ExecutionError(f"Local tap forward {local_fwd.pid} died with {local_fwd.returncode}")

                time.sleep(0.1)

            yield ns



"""Interface for controlling relays of Shelly devices using the Gen 1 API

NetworkPowerPort:
    model: shelly_gen1
    host: 'http://192.168.0.42'
    index: 0

Will do a GET request to http://192.168.0.42/relay/0 to get the current
relay state, and a POST request to http://192.168.0.42/relay/0 with request
data of 'turn=off' or 'turn=on' to change the relay state.

Also, see the official Gen 1 Device API documentation:
https://shelly-api-docs.shelly.cloud/gen1/

"""

import json

import requests


def power_set(host: str, port: int, index: int = 0, value: bool = True):
    assert not port
    turn = "on" if value else "off"
    r = requests.post(f"{host}/relay/{index}", data={"turn": turn})
    r.raise_for_status()


def power_get(host: str, port: int, index: int):
    assert not port
    r = requests.get(f"{host}/relay/{index}")
    r.raise_for_status()
    return json.loads(r.text)["ison"]



import attr

from .common import Driver
from ..factory import target_factory
from ..step import step
from .exception import ExecutionError
from ..util.helper import processwrapper


@target_factory.reg_driver
@attr.s(eq=False)
class USBSDWireDriver(Driver):
    """The USBSDWireDriver uses the sd-mux-ctrl tool to control SDWire hardware

    Args:
        bindings (dict): driver to use with usbsdmux
    """

    bindings = {
        "mux": {"USBSDWireDevice", "NetworkUSBSDWireDevice"},
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.target.env:
            self.tool = self.target.env.config.get_tool("sd-mux-ctrl")
        else:
            self.tool = "sd-mux-ctrl"

    @Driver.check_active
    @step(title="sdmux_set", args=["mode"])
    def set_mode(self, mode):
        if not mode.lower() in ["dut", "host"]:
            raise ExecutionError(f"Setting mode '{mode}' not supported by USBSDWireDriver")
        cmd = self.mux.command_prefix + [
            self.tool,
            "--dut" if mode.lower() == "dut" else "--ts",
            "-e",
            self.mux.control_serial,
        ]
        processwrapper.check_output(cmd)

    @Driver.check_active
    @step(title="sdmux_get")
    def get_mode(self):
        cmd = self.mux.command_prefix + [
            self.tool,
            "-e",
            self.mux.control_serial,
            "-u",
        ]
        result = processwrapper.check_output(cmd)
        return result.split(b": ", maxsplit=1)[1].strip().decode()



import functools

import attr

from ..binding import BindingError
from ..driver import Driver


def never_retry(func):
    """
    Strategy method decorator storing the original exception in strategy.broken and raising a
    StrategyError from it. All subsequent calls to the decorated method will lead to the same
    exception.
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.broken:
            # a previous call broke the strategy, raise the original exception again
            raise StrategyError(f"{self.__class__.__name__} is in broken state") from self.broken

        try:
            return func(self, *args, **kwargs)
        except Exception as e:
            # store original exception and raise StrategyError from it
            self.broken = e
            raise StrategyError(f"{self.__class__.__name__} is in broken state") from self.broken

    return wrapper


@attr.s(eq=False)
class StrategyError(Exception):
    msg = attr.ib(validator=attr.validators.instance_of(str))


@attr.s(eq=False)
class Strategy(Driver):  # reuse driver handling
    """
    Represents a strategy which places a target into a requested state by
    calling specific drivers. A strategy usually needs to know some details of
    a given target.

    Life cycle:
    - create
    - bind (n times)
    - usage

    TODO: This might also be just a driver?
    """

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        # contains exception that broke the strategy or None
        self.broken = None

        if self.target is None:
            raise BindingError("Strategies can only be created on a valid target")

    def on_client_bound(self, client):
        raise NotImplementedError("Strategies do not support clients")

    def on_activate(self):
        raise NotImplementedError("Strategies can not be activated")

    def on_deactivate(self):
        pass

    def resolve_conflicts(self, client):
        raise NotImplementedError("Strategies do not support clients")

    def transition(self, status):
        raise NotImplementedError(f"Strategy.transition() is not implemented for {self.__class__.__name__}")

    def force(self, status):
        raise NotImplementedError(f"Strategy.force() is not implemented for {self.__class__.__name__}")

    def prepare_export(self):
        """By default, export all drivers bound by the strategy."""
        name_map = {}
        for name in self.bindings.keys():
            name_map[getattr(self, name)] = name
        return name_map



import requests

# Driver has been tested with:
# Gude  Expert Power Control 8225-1 - v1.0.6

# HTTP-GET API is defined in the Gude EPC-HTTP-Interface specification:
# http://wiki.gude.info/EPC_HTTP_Interface

PORT = 80


def power_set(host, port, index, value):
    index = int(index)
    assert 1 <= index <= 12

    value = 1 if value else 0
    r = requests.get(f"http://{host}:{port}/ov.html?cmd=1&p={index}&s={value}")
    r.raise_for_status()


def power_get(host, port, index):
    index = int(index)
    assert 1 <= index <= 12

    r = requests.get(f"http://{host}:{port}/statusjsn.js?components=1")
    r.raise_for_status()

    state = r.json()["outputs"][index - 1]["state"]

    return state



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import labgrid_coordinator_pb2 as labgrid__coordinator__pb2


class CoordinatorStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ClientStream = channel.stream_stream(
                '/labgrid.Coordinator/ClientStream',
                request_serializer=labgrid__coordinator__pb2.ClientInMessage.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.ClientOutMessage.FromString,
                )
        self.ExporterStream = channel.stream_stream(
                '/labgrid.Coordinator/ExporterStream',
                request_serializer=labgrid__coordinator__pb2.ExporterInMessage.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.ExporterOutMessage.FromString,
                )
        self.AddPlace = channel.unary_unary(
                '/labgrid.Coordinator/AddPlace',
                request_serializer=labgrid__coordinator__pb2.AddPlaceRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.AddPlaceResponse.FromString,
                )
        self.DeletePlace = channel.unary_unary(
                '/labgrid.Coordinator/DeletePlace',
                request_serializer=labgrid__coordinator__pb2.DeletePlaceRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.DeletePlaceResponse.FromString,
                )
        self.GetPlaces = channel.unary_unary(
                '/labgrid.Coordinator/GetPlaces',
                request_serializer=labgrid__coordinator__pb2.GetPlacesRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.GetPlacesResponse.FromString,
                )
        self.AddPlaceAlias = channel.unary_unary(
                '/labgrid.Coordinator/AddPlaceAlias',
                request_serializer=labgrid__coordinator__pb2.AddPlaceAliasRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.AddPlaceAliasResponse.FromString,
                )
        self.DeletePlaceAlias = channel.unary_unary(
                '/labgrid.Coordinator/DeletePlaceAlias',
                request_serializer=labgrid__coordinator__pb2.DeletePlaceAliasRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.DeletePlaceAliasResponse.FromString,
                )
        self.SetPlaceTags = channel.unary_unary(
                '/labgrid.Coordinator/SetPlaceTags',
                request_serializer=labgrid__coordinator__pb2.SetPlaceTagsRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.SetPlaceTagsResponse.FromString,
                )
        self.SetPlaceComment = channel.unary_unary(
                '/labgrid.Coordinator/SetPlaceComment',
                request_serializer=labgrid__coordinator__pb2.SetPlaceCommentRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.SetPlaceCommentResponse.FromString,
                )
        self.AddPlaceMatch = channel.unary_unary(
                '/labgrid.Coordinator/AddPlaceMatch',
                request_serializer=labgrid__coordinator__pb2.AddPlaceMatchRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.AddPlaceMatchResponse.FromString,
                )
        self.DeletePlaceMatch = channel.unary_unary(
                '/labgrid.Coordinator/DeletePlaceMatch',
                request_serializer=labgrid__coordinator__pb2.DeletePlaceMatchRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.DeletePlaceMatchResponse.FromString,
                )
        self.AcquirePlace = channel.unary_unary(
                '/labgrid.Coordinator/AcquirePlace',
                request_serializer=labgrid__coordinator__pb2.AcquirePlaceRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.AcquirePlaceResponse.FromString,
                )
        self.ReleasePlace = channel.unary_unary(
                '/labgrid.Coordinator/ReleasePlace',
                request_serializer=labgrid__coordinator__pb2.ReleasePlaceRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.ReleasePlaceResponse.FromString,
                )
        self.AllowPlace = channel.unary_unary(
                '/labgrid.Coordinator/AllowPlace',
                request_serializer=labgrid__coordinator__pb2.AllowPlaceRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.AllowPlaceResponse.FromString,
                )
        self.CreateReservation = channel.unary_unary(
                '/labgrid.Coordinator/CreateReservation',
                request_serializer=labgrid__coordinator__pb2.CreateReservationRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.CreateReservationResponse.FromString,
                )
        self.CancelReservation = channel.unary_unary(
                '/labgrid.Coordinator/CancelReservation',
                request_serializer=labgrid__coordinator__pb2.CancelReservationRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.CancelReservationResponse.FromString,
                )
        self.PollReservation = channel.unary_unary(
                '/labgrid.Coordinator/PollReservation',
                request_serializer=labgrid__coordinator__pb2.PollReservationRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.PollReservationResponse.FromString,
                )
        self.GetReservations = channel.unary_unary(
                '/labgrid.Coordinator/GetReservations',
                request_serializer=labgrid__coordinator__pb2.GetReservationsRequest.SerializeToString,
                response_deserializer=labgrid__coordinator__pb2.GetReservationsResponse.FromString,
                )


class CoordinatorServicer(object):
    """Missing associated documentation comment in .proto file."""

    def ClientStream(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ExporterStream(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddPlace(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeletePlace(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetPlaces(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddPlaceAlias(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeletePlaceAlias(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetPlaceTags(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetPlaceComment(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddPlaceMatch(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeletePlaceMatch(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AcquirePlace(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReleasePlace(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AllowPlace(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateReservation(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CancelReservation(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PollReservation(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetReservations(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_CoordinatorServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ClientStream': grpc.stream_stream_rpc_method_handler(
                    servicer.ClientStream,
                    request_deserializer=labgrid__coordinator__pb2.ClientInMessage.FromString,
                    response_serializer=labgrid__coordinator__pb2.ClientOutMessage.SerializeToString,
            ),
            'ExporterStream': grpc.stream_stream_rpc_method_handler(
                    servicer.ExporterStream,
                    request_deserializer=labgrid__coordinator__pb2.ExporterInMessage.FromString,
                    response_serializer=labgrid__coordinator__pb2.ExporterOutMessage.SerializeToString,
            ),
            'AddPlace': grpc.unary_unary_rpc_method_handler(
                    servicer.AddPlace,
                    request_deserializer=labgrid__coordinator__pb2.AddPlaceRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.AddPlaceResponse.SerializeToString,
            ),
            'DeletePlace': grpc.unary_unary_rpc_method_handler(
                    servicer.DeletePlace,
                    request_deserializer=labgrid__coordinator__pb2.DeletePlaceRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.DeletePlaceResponse.SerializeToString,
            ),
            'GetPlaces': grpc.unary_unary_rpc_method_handler(
                    servicer.GetPlaces,
                    request_deserializer=labgrid__coordinator__pb2.GetPlacesRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.GetPlacesResponse.SerializeToString,
            ),
            'AddPlaceAlias': grpc.unary_unary_rpc_method_handler(
                    servicer.AddPlaceAlias,
                    request_deserializer=labgrid__coordinator__pb2.AddPlaceAliasRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.AddPlaceAliasResponse.SerializeToString,
            ),
            'DeletePlaceAlias': grpc.unary_unary_rpc_method_handler(
                    servicer.DeletePlaceAlias,
                    request_deserializer=labgrid__coordinator__pb2.DeletePlaceAliasRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.DeletePlaceAliasResponse.SerializeToString,
            ),
            'SetPlaceTags': grpc.unary_unary_rpc_method_handler(
                    servicer.SetPlaceTags,
                    request_deserializer=labgrid__coordinator__pb2.SetPlaceTagsRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.SetPlaceTagsResponse.SerializeToString,
            ),
            'SetPlaceComment': grpc.unary_unary_rpc_method_handler(
                    servicer.SetPlaceComment,
                    request_deserializer=labgrid__coordinator__pb2.SetPlaceCommentRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.SetPlaceCommentResponse.SerializeToString,
            ),
            'AddPlaceMatch': grpc.unary_unary_rpc_method_handler(
                    servicer.AddPlaceMatch,
                    request_deserializer=labgrid__coordinator__pb2.AddPlaceMatchRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.AddPlaceMatchResponse.SerializeToString,
            ),
            'DeletePlaceMatch': grpc.unary_unary_rpc_method_handler(
                    servicer.DeletePlaceMatch,
                    request_deserializer=labgrid__coordinator__pb2.DeletePlaceMatchRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.DeletePlaceMatchResponse.SerializeToString,
            ),
            'AcquirePlace': grpc.unary_unary_rpc_method_handler(
                    servicer.AcquirePlace,
                    request_deserializer=labgrid__coordinator__pb2.AcquirePlaceRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.AcquirePlaceResponse.SerializeToString,
            ),
            'ReleasePlace': grpc.unary_unary_rpc_method_handler(
                    servicer.ReleasePlace,
                    request_deserializer=labgrid__coordinator__pb2.ReleasePlaceRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.ReleasePlaceResponse.SerializeToString,
            ),
            'AllowPlace': grpc.unary_unary_rpc_method_handler(
                    servicer.AllowPlace,
                    request_deserializer=labgrid__coordinator__pb2.AllowPlaceRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.AllowPlaceResponse.SerializeToString,
            ),
            'CreateReservation': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateReservation,
                    request_deserializer=labgrid__coordinator__pb2.CreateReservationRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.CreateReservationResponse.SerializeToString,
            ),
            'CancelReservation': grpc.unary_unary_rpc_method_handler(
                    servicer.CancelReservation,
                    request_deserializer=labgrid__coordinator__pb2.CancelReservationRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.CancelReservationResponse.SerializeToString,
            ),
            'PollReservation': grpc.unary_unary_rpc_method_handler(
                    servicer.PollReservation,
                    request_deserializer=labgrid__coordinator__pb2.PollReservationRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.PollReservationResponse.SerializeToString,
            ),
            'GetReservations': grpc.unary_unary_rpc_method_handler(
                    servicer.GetReservations,
                    request_deserializer=labgrid__coordinator__pb2.GetReservationsRequest.FromString,
                    response_serializer=labgrid__coordinator__pb2.GetReservationsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'labgrid.Coordinator', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Coordinator(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def ClientStream(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/labgrid.Coordinator/ClientStream',
            labgrid__coordinator__pb2.ClientInMessage.SerializeToString,
            labgrid__coordinator__pb2.ClientOutMessage.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ExporterStream(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/labgrid.Coordinator/ExporterStream',
            labgrid__coordinator__pb2.ExporterInMessage.SerializeToString,
            labgrid__coordinator__pb2.ExporterOutMessage.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AddPlace(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/AddPlace',
            labgrid__coordinator__pb2.AddPlaceRequest.SerializeToString,
            labgrid__coordinator__pb2.AddPlaceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeletePlace(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/DeletePlace',
            labgrid__coordinator__pb2.DeletePlaceRequest.SerializeToString,
            labgrid__coordinator__pb2.DeletePlaceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetPlaces(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/GetPlaces',
            labgrid__coordinator__pb2.GetPlacesRequest.SerializeToString,
            labgrid__coordinator__pb2.GetPlacesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AddPlaceAlias(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/AddPlaceAlias',
            labgrid__coordinator__pb2.AddPlaceAliasRequest.SerializeToString,
            labgrid__coordinator__pb2.AddPlaceAliasResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeletePlaceAlias(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/DeletePlaceAlias',
            labgrid__coordinator__pb2.DeletePlaceAliasRequest.SerializeToString,
            labgrid__coordinator__pb2.DeletePlaceAliasResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetPlaceTags(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/SetPlaceTags',
            labgrid__coordinator__pb2.SetPlaceTagsRequest.SerializeToString,
            labgrid__coordinator__pb2.SetPlaceTagsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetPlaceComment(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/SetPlaceComment',
            labgrid__coordinator__pb2.SetPlaceCommentRequest.SerializeToString,
            labgrid__coordinator__pb2.SetPlaceCommentResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AddPlaceMatch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/AddPlaceMatch',
            labgrid__coordinator__pb2.AddPlaceMatchRequest.SerializeToString,
            labgrid__coordinator__pb2.AddPlaceMatchResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeletePlaceMatch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/DeletePlaceMatch',
            labgrid__coordinator__pb2.DeletePlaceMatchRequest.SerializeToString,
            labgrid__coordinator__pb2.DeletePlaceMatchResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AcquirePlace(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/AcquirePlace',
            labgrid__coordinator__pb2.AcquirePlaceRequest.SerializeToString,
            labgrid__coordinator__pb2.AcquirePlaceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ReleasePlace(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/ReleasePlace',
            labgrid__coordinator__pb2.ReleasePlaceRequest.SerializeToString,
            labgrid__coordinator__pb2.ReleasePlaceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AllowPlace(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/AllowPlace',
            labgrid__coordinator__pb2.AllowPlaceRequest.SerializeToString,
            labgrid__coordinator__pb2.AllowPlaceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateReservation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/CreateReservation',
            labgrid__coordinator__pb2.CreateReservationRequest.SerializeToString,
            labgrid__coordinator__pb2.CreateReservationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CancelReservation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/CancelReservation',
            labgrid__coordinator__pb2.CancelReservationRequest.SerializeToString,
            labgrid__coordinator__pb2.CancelReservationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PollReservation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/PollReservation',
            labgrid__coordinator__pb2.PollReservationRequest.SerializeToString,
            labgrid__coordinator__pb2.PollReservationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetReservations(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/labgrid.Coordinator/GetReservations',
            labgrid__coordinator__pb2.GetReservationsRequest.SerializeToString,
            labgrid__coordinator__pb2.GetReservationsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import attr

from ..factory import target_factory
from ..util import gen_marker
from ..step import step
from .common import Driver

from .ubootdriver import UBootDriver
from ..util import re_vt100

@target_factory.reg_driver
@attr.s(eq=False)
class SmallUBootDriver(UBootDriver):
    """
    SmallUBootDriver is meant as a driver for UBoot with only little
    functionality compared to standard a standard UBoot.
    Especially is copes with the following limitations:

    - The UBoot does not have a real password-prompt but can be activated by
      entering a "secret" after a message was displayed.
    - The command line is does not have a build-in echo command. Thus this
      driver uses 'Unknown Command' messages as marker before and after the
      output of a command.
    - Since there is no echo we can not return the exit code of the command.
      Commands will always return 0 unless the command was not found.

    This driver needs the following features activated in UBoot to work:

    - The UBoot must not have real password prompt. Instead it must be
      keyword activated.
      For example it should be activated by a dialog like the following:
      UBoot: "Autobooting in 1s..."
      Labgrid: "secret"
      UBoot: <switching to console>
    - The UBoot must be able to parse multiple commands in a single
      line separated by ";".
    - The UBoot must support the "bootm" command to boot from a
      memory location.

    This driver was created especially for the following devices:

    - TP-Link WR841 v11

    Args:
        boot_secret (str): optional, secret used to unlock prompt
        boot_secret_nolf (bool): optional, send boot_secret without new line
        login_timeout (int): optional, timeout for login prompt detection,
    """

    boot_expression = attr.ib(default=r"U-Boot 20\d+", validator=attr.validators.instance_of(str))
    boot_secret = attr.ib(default="a", validator=attr.validators.instance_of(str))
    boot_secret_nolf = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    login_timeout = attr.ib(default=60, validator=attr.validators.instance_of(int))

    @step()
    def _await_prompt(self):
        """
        Await autoboot_expression. If this line was read enter the 'secret' (or a
        single character) to interrupt normal boot.
        """

        # wait for boot expression. Afterwards enter secret
        self.console.expect(self.boot_expression, timeout=self.login_timeout)
        if self.boot_secret_nolf:
            self.console.write(self.boot_secret.encode('ASCII'))
        else:
            self.console.sendline(self.boot_secret)
        self._status = 1

        # wait until UBoot has reached it's prompt
        self.console.expect(self.prompt)
        for command in self.init_commands:
            self._run(command)

    def _run(self, cmd: str, *, timeout: int = 30, codec: str = "utf-8", decodeerrors: str = "strict"):  # pylint: disable=line-too-long
        """
        If Uboot is in Command-Line mode: Run command cmd and return it's
        output.

        Arguments:
        cmd - Command to run
        """
        # TODO: use codec, decodeerrors

        # Check if Uboot is in command line mode
        if self._status != 1:
            return None

        marker = gen_marker()

        # Create multi-part command like we would do for a normal uboot.
        # but since this simple uboot does not have an echo-command we will
        # handle it's error message as an echo-output.
        # additionally we are not able to get the command's return code and
        # will always return 0.
        cmp_command = f"echo{marker}; {cmd}; echo{marker}"

        self.console.sendline(cmp_command)
        _, before, _, _ = self.console.expect(self.prompt, timeout=timeout)

        data = re_vt100.sub(
            '', before.decode('utf-8'), count=1000000
        ).replace("\r", "").split("\n")
        data = data[1:]
        data = data[data.index(f"Unknown command 'echo{marker}' - try 'help'") +1 :]
        data = data[:data.index(f"Unknown command 'echo{marker}' - try 'help'")]
        if len(data) >= 1:
            if data[0].startswith("Unknown command '"):
                return (data, [], 1)
        return (data, [], 0)

    @Driver.check_active
    @step(args=['name'], result=True)
    def boot(self, name):
        """
        Boot the device from the given memory location using 'bootm'.

        Args:
            name (str): address to boot
        """
        self.console.sendline(f"bootm {name}")



from tempfile import TemporaryDirectory
import os

from .common import Strategy, StrategyError
from ..step import step

__all__ = [
    "InvalidGraphStrategyError",
    "GraphStrategyRuntimeError",
    "GraphStrategyError",
    "GraphStrategy",
]


class GraphStrategyError(StrategyError):
    """Generic GraphStrategy error"""

    pass


class InvalidGraphStrategyError(GraphStrategyError):
    """GraphStrategy error raised during initialization of broken strategies"""

    pass


class GraphStrategyRuntimeError(GraphStrategyError):
    """GraphStrategy error raised during runtime when used in unintended ways"""

    pass


class GraphStrategy(Strategy):
    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self.__transition_running = False

        # find states
        self.states = {}

        for state_name in dir(self):
            if not state_name.startswith("state_"):
                continue

            method = getattr(self, state_name)

            if not callable(method):
                raise InvalidGraphStrategyError(f"GraphStrategy state '{state_name}' is not callable")

            state_name = "_".join(state_name.split("_")[1:])

            self.states[state_name] = {
                "method": step()(method),
                "dependencies": getattr(method, "dependencies", []),
            }

        if not self.states:
            raise InvalidGraphStrategyError("GraphStrategies without states are invalid")

        # check dependencies
        state_names = self.states.keys()

        for state_name in state_names:
            for dependency in self.states[state_name]["dependencies"]:
                if dependency not in state_names:
                    raise InvalidGraphStrategyError(
                        f"{state_name}: State '{dependency}' is unknown. State names are: {', '.join(state_names)}"  # pylint: disable=line-too-long
                    )

        # find root state
        root_states = [k for k, v in self.states.items() if not v["dependencies"]]

        if not root_states:
            raise InvalidGraphStrategyError("GraphStrategies without root state are invalid")

        # check check if exact one root state is defined
        if len(root_states) > 1:
            raise InvalidGraphStrategyError(
                "Only one root state supported. Defined root states: {}".format(  # NOQA
                    ", ".join(root_states),
                )
            )

        self.root_state = root_states[0]
        self.invalidate()

        # setup grahviz cache
        self._graph_cache = {
            "tempdir": None,
            "graph": None,
            "path": self.path,
        }

    def invalidate(self):
        """
        Marks the path to the current state as out-of-date. Subsequent transition() calls will
        start from the root state.
        Will be called if exceptions in state methods occur.
        """
        self.path = []

        # deactivate all drivers to restore initial state
        self.target.deactivate_all_drivers()

    @step(args=["state"])
    def transition(self, state, via=None):
        """
        Computes the path from root state (via "via" state, if given) to given state.
        If the computed path is fully incremental to the path executed previously, only the state's
        methods relative to the previous path are executed. Otherwise all states' methods of the
        computed path (starting from the root node) are executed.
        """
        if not isinstance(via, (type(None), list)):
            raise GraphStrategyRuntimeError("'via' has to be a list or None")
        # for use with labgrid-client -s, if only state is set, try to extract
        # the via states
        if ":" in state and via is None:
            state, via = state.split(":")
            via = via.split(",")
        via = via or []
        try:
            # check if another transition is running
            if self.__transition_running:
                raise GraphStrategyRuntimeError("Another transition is already running")

            # lock transition
            self.__transition_running = True

            # check if state is known
            if state not in self.states:
                raise GraphStrategyRuntimeError(
                    f"Unknown state '{state}'. State names are: {', '.join(self.states.keys())}"
                )

            # find path
            abs_path = self.find_abs_path(state, via=via)

            if abs_path == self.path:
                return []

            path = self.find_rel_path(abs_path)

            # run state methods
            for state_name in path:
                if state_name == self.root_state:
                    # deactivate drivers before root state method is called
                    self.target.deactivate_all_drivers()

                try:
                    self.states[state_name]["method"]()

                except Exception:
                    self.invalidate()

                    raise

            self.path = abs_path

            return path

        finally:
            # unlock transition
            self.__transition_running = False

    def find_abs_path(self, state, via=None):
        """
        Computes the absolute path from the root state, via "via" (if given), to the given state.
        """
        via = via or []
        via = via[::-1]
        path = [
            state,
        ]
        current_state = self.states[state]

        for via_state in via:
            if via_state not in self.states.keys():
                raise GraphStrategyRuntimeError(
                    f"Unknown state '{via_state}' in via. State names are: {', '.join(self.states.keys())}"  # pylint: disable=line-too-long
                )

        while current_state["dependencies"]:
            next_state = current_state["dependencies"][0]

            for i in via:
                if i in current_state["dependencies"]:
                    via.remove(i)
                    next_state = i

            path.insert(0, next_state)
            current_state = self.states[next_state]

        # no via states should be left now
        if via:
            raise GraphStrategyRuntimeError(
                "Path to '{}' via {} does not exist".format(state, ", ".join(["'{}'".format(v) for v in via]))
            )

        return path

    def find_rel_path(self, path):
        """
        If the given path is fully incremental to the path executed before, returns the path
        relative to the previously executed one.
        Otherwise the given path is returned.
        """
        if path[: -(len(path) - len(self.path))] == self.path:
            return path[len(self.path) :]

        return path

    @property
    def graph(self):
        """
        Returns a graphviz.Digraph for the directed graph the inheriting strategy represents.

        The graph can be rendered with:
        ``mystrategy.graph.render("filename") # renders to filename.png``
        """
        from graphviz import Digraph

        if self._graph_cache["graph"] and self._graph_cache["path"] == self.path:
            return self._graph_cache["graph"]

        if not self._graph_cache["tempdir"]:
            self._graph_cache["tempdir"] = TemporaryDirectory()

        dg = Digraph(
            filename=os.path.join(self._graph_cache["tempdir"].name, "graph"),
            format="png",
        )

        edges = []

        dg.attr("node", style="filled", fillcolor="lightblue2", penwidth="1")
        dg.attr("edge", style="solid")

        for index, node_name in enumerate(self.path):
            attrs = {}

            if node_name == self.path[-1]:
                attrs = {"penwidth": "2"}

            dg.node(node_name, **attrs)

            if index < len(self.path) - 1:
                edges.append(
                    (
                        node_name,
                        self.path[index + 1],
                    )
                )
                dg.edge(*edges[-1])

        dg.attr("node", style="filled", color="lightgrey", fillcolor="lightgrey")

        dg.attr("edge", style="dashed", arrowhead="empty")

        for node_name in self.states:
            if node_name not in self.path:
                dg.node(node_name)

            for edge in self.states[node_name]["dependencies"]:
                if (
                    edge,
                    node_name,
                ) in edges:
                    continue

                dg.edge(edge, node_name)

        self._graph_cache["graph"] = dg

        return dg

    @classmethod
    def depends(cls, *dependencies):
        """``@depends`` decorator used to list states the decorated state directly depends on."""

        def decorator(function):
            function.dependencies = list(dependencies)

            return function

        return decorator



import requests

from ..exception import ExecutionError


# This driver implements a power port for Gude EPC-8316 Power Switches.
# These switches differ in their API to the previous switches for
# get-commands (because of a firmware bug).
#
# Apart from a workaround for this bug and reduced port index limits of
# 8 instead of 24, this implementation works just like gude24.py.
#
# Driver has been tested with:
# * Gude Expert Power Control 8316 (firmware v1.2.0)

PORT = 80


def power_set(host, port, index, value):
    # The gude web-interface uses different pages for the three groups of
    # switches. The web-interface always uses the 'correct' page to set a
    # value. But commands for all pages are accepted on all pages.
    index = int(index)
    assert 1 <= index <= 8
    # access the web interface...
    value = 1 if value else 0
    r = requests.get(f"http://{host}:{port}/ov.html?cmd=1&p={index}&s={value}")
    r.raise_for_status()


def power_get(host, port, index):
    # The status of the ports is made available via a html <meta>-tag using the
    # following format:
    # <meta http-equiv="powerstate" content="Power Port ,0">
    #
    # The index of each port should normally be in there, but isn't, because of
    # a firmware bug:
    # <meta http-equiv="powerstate" content="Power Port 1,0">
    #
    # Again the status of all ports is made available on all pages.
    index = int(index)
    assert 1 <= index <= 8
    # get the contents of the main page
    r = requests.get(f"http://{host}:{port}/ov.html")
    r.raise_for_status()
    for line_no, line in enumerate(r.text.splitlines()):
        if line_no == index and line.find('content="Power Port ') > 0:
            if line.find(",0") > 0:
                return False
            if line.find(",1") > 0:
                return True

            raise ExecutionError("failed to parse the port status")
    # if we got this far, something is wrong with the website
    raise ExecutionError("failed to find the port")



"""
This driver implements a power port for Gude Power Switches with up to
24 ports.
These switches differ in their API to the previous 8-port switches for set-
and get-commands.

Driver has been tested with:
* Gude Expert Power Control 8080
"""

import re

import requests

from ..exception import ExecutionError


PORT = 80


def power_set(host, port, index, value):
    """
    The gude web-interface uses different pages for the three groups of
    switches. The web-interface always uses the 'correct' page to set a
    value. But commands for all pages are accepted on all pages.
    """
    index = int(index)
    assert 1 <= index <= 24
    # access the web interface...
    value = 1 if value else 0
    response = requests.get(f"http://{host}:{port}/ov.html?cmd=1&p={index}&s={value}")

    # Check, that the port is in the desired state
    state = get_state(response, index)
    if state != value:
        raise ExecutionError(f"failed to set port {index} to status {value}")


def power_get(host, port, index):
    """
    Get the status of a port.
    """
    index = int(index)
    assert 1 <= index <= 24
    # get the contents of the main page
    response = requests.get(f"http://{host}:{port}/ov.html")
    state = get_state(response, index)
    return state


def get_state(request, index):
    """
    The status of the ports is made available via a html <meta>-tag using the
    following format:
    <meta http-equiv="powerstate" content="Power Port 1,0">
    Again the status of all ports is made available on all pages.
    """
    request.raise_for_status()

    # Get the power state of a specified index
    # raise an exception if the state cannot be determined.

    for line in request.text.splitlines():
        if line.find(f'content="Power Port {index}') > 0:
            if line.find(",0") > 0:
                return False
            if line.find(",1") > 0:
                return True

            raise ExecutionError("failed to parse the port status")

    m = re.search(r"Device is blocked by another user with IP ([^\s<]+)", request.text)
    if m:
        raise ExecutionError(f"device is blocked by another user with IP {m.group(1)}")

    # if we got this far, something is wrong with the website
    raise ExecutionError(f"failed to determine status of power port {index}")



import attr
from pexpect import TIMEOUT
import serial
import serial.rfc2217

from ..factory import target_factory
from ..protocol import ConsoleProtocol
from .common import Driver
from .consoleexpectmixin import ConsoleExpectMixin
from ..util.proxy import proxymanager
from ..resource import SerialPort


@target_factory.reg_driver
@attr.s(eq=False)
class SerialDriver(ConsoleExpectMixin, Driver, ConsoleProtocol):
    """
    Driver implementing the ConsoleProtocol interface over a SerialPort connection
    """
    bindings = {"port": {"SerialPort", "NetworkSerialPort"}, }

    txdelay = attr.ib(default=0.0, validator=attr.validators.instance_of(float))
    txchunk = attr.ib(default=1, validator=attr.validators.instance_of(int))
    timeout = attr.ib(default=3.0, validator=attr.validators.instance_of(float))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if isinstance(self.port, SerialPort):
            self.serial = serial.Serial()
        else:
            if self.port.protocol == "rfc2217":
                self.serial = serial.rfc2217.Serial()
            elif self.port.protocol == "raw":
                self.serial = serial.serial_for_url("socket://", do_not_open=True)
            else:
                raise Exception("SerialDriver: unknown protocol")
        self.status = 0

    def on_activate(self):
        if isinstance(self.port, SerialPort):
            self.serial.port = self.port.port
            self.serial.baudrate = self.port.speed
        else:
            host, port = proxymanager.get_host_and_port(self.port)
            if self.port.protocol == "rfc2217":
                self.serial.port = f"rfc2217://{host}:{port}?ign_set_control&timeout={self.timeout}"
            elif self.port.protocol == "raw":
                self.serial.port = f"socket://{host}:{port}/"
            else:
                raise Exception("SerialDriver: unknown protocol")
            self.serial.baudrate = self.port.speed
        self.open()

    def on_deactivate(self):
        self.close()

    @Driver.check_bound
    def get_export_vars(self):
        export_vars = {
            "speed": str(self.port.speed)
        }
        if isinstance(self.port, SerialPort):
            export_vars["port"] = self.port.port
        else:
            host, port = proxymanager.get_host_and_port(self.port)
            export_vars["host"] = host
            export_vars["port"] = str(port)
            export_vars["protocol"] = self.port.protocol
        return export_vars

    def _read(self, size: int = 1, timeout: float = 0.0, max_size: int = None):
        """
        Reads 'size' or more bytes from the serialport

        Keyword Arguments:
        size -- amount of bytes to read, defaults to 1
        max_size -- maximal amount of bytes to read, values 'None' or '0' do not restrict the read
                    length, defaults to None
        if size == max_size: read and return exactly size = max_size bytes
        """
        reading = max(size, self.serial.in_waiting)
        if max_size:  # limit reading to max_size if provided
            reading = min(reading, max_size)
        self.serial.timeout = timeout
        res = self.serial.read(reading)
        if not res:
            raise TIMEOUT(f"Timeout of {timeout:.2f} seconds exceeded or connection closed by peer")
        return res

    def _write(self, data: bytes):
        """
        Writes 'data' to the serialport

        Arguments:
        data -- data to write, must be bytes
        """
        return self.serial.write(data)

    def open(self):
        """Opens the serialport, does nothing if it is already open"""
        if not self.status:
            try:
                self.serial.open()
            except serial.SerialException as e:
                raise serial.SerialException(
                    f"Could not open serial port {self.serial.port}: {str(e)}") from e

            self.status = 1

    def close(self):
        """Closes the serialport, does nothing if it is already closed"""
        if self.status:
            self.serial.close()
            self.status = 0

    def __str__(self):
        return f"SerialDriver({self.target.name})"



"""Simple rest interface for Power Port. Used for ex. misc Raspberry Pi configs
  Author: Kjeld Flarup <kfa@deif.com>

  The URL given in hosts in exporter.yaml must replace {value} with '0' or '1'
  It is optional whether to use {index} or not.

NetworkPowerPort:
    model: simplerest
    host: 'http://172.17.180.53:9999/relay/{index}/{value}'
    index: 0
"""

import requests


def power_set(host, port, index, value):
    assert port is None

    index = int(index)
    value = 1 if value else 0
    r = requests.get(host.format(value=value, index=index))
    r.raise_for_status()


def power_get(host, port, index):
    assert port is None

    index = int(index)
    # remove trailing /
    r = requests.get(host.format(value="", index=index).rstrip("/"))
    r.raise_for_status()
    return r.text == "1"



import abc


class FileSystemProtocol(abc.ABC):
    @abc.abstractmethod
    def read(self, filename: str):
        raise NotImplementedError

    @abc.abstractmethod
    def write(self, filename: str, data: bytes, append: bool):
        raise NotImplementedError



"""Driver for the Raritan PDUs

Documentation sources:
* https://www.circitor.fr/Mibs/Html/P/PDU2-MIB.php
* http://support.raritan.com/px3/version-3.0.3/user-guides/PX3-0B-v3.0-E.pdf
* https://cdn1.raritan.com/download/pdu-g2/4.0.20/MIB_Usage_4.0.20_49038.pdf
"""
from ..exception import ExecutionError
from ...util.snmp import SimpleSNMP


OID = ".1.3.6.1.4.1.13742.6.4.1.2.1"


def power_set(host, port, index, value):
    _snmp = SimpleSNMP(host, 'private', port=port)
    outlet_control_oid = "{}.2.1.{}".format(OID, index)

    _snmp.set(outlet_control_oid, str(int(value)))


def power_get(host, port, index):
    _snmp = SimpleSNMP(host, 'public', port=port)
    output_status_oid = "{}.3.1.{}".format(OID, index)

    value = _snmp.get(output_status_oid)

    if value == 7:  # On
        return True
    if value == 8:  # Off
        return False

    raise ExecutionError("failed to get SNMP value")



import time

import attr

from ..factory import target_factory
from ..protocol import DigitalOutputProtocol, ResetProtocol
from ..step import step
from .common import Driver

@target_factory.reg_driver
@attr.s(eq=False)
class DigitalOutputResetDriver(Driver, ResetProtocol):
    """DigitalOutputResetDriver - Driver using a DigitalOutput to reset the
    target"""
    bindings = {"output": DigitalOutputProtocol, }
    delay = attr.ib(default=1.0, validator=attr.validators.instance_of(float))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @Driver.check_active
    @step()
    def reset(self):
        self.output.set(True)
        time.sleep(self.delay)
        self.output.set(False)



from pprint import pprint

import glob
import os

import attr
import jinja2

from ..util.yaml import load
from ..exceptions import NoConfigFoundError


@attr.s(eq=False)
class ResourceConfig:
    filename = attr.ib(validator=attr.validators.instance_of(str))
    template_env = attr.ib(default=attr.Factory(dict), validator=attr.validators.instance_of(dict))

    def __attrs_post_init__(self):
        _dirname = os.path.dirname(self.filename)
        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(_dirname),
            line_statement_prefix="#",
            line_comment_prefix="##",
        )
        try:
            template = env.get_template(os.path.basename(self.filename))
        except jinja2.TemplateNotFound:
            raise NoConfigFoundError(f"{self.filename} could not be found")
        drop_ins = [
            os.path.relpath(p, _dirname) for p in sorted(glob.glob(os.path.join(self.filename + ".d", "*.yaml")))
        ]

        self.template_env["drop_ins"] = drop_ins
        rendered = template.render(self.template_env)
        pprint(("rendered", rendered))
        self.data = load(rendered)
        pprint(("loaded", self.data))



import fcntl
import os
import select
import shlex
import subprocess

import attr

from ..factory import target_factory
from ..protocol import ConsoleProtocol
from .common import Driver
from .consoleexpectmixin import ConsoleExpectMixin
from .exception import ExecutionError


@target_factory.reg_driver
@attr.s(eq=False)
class ExternalConsoleDriver(ConsoleExpectMixin, Driver, ConsoleProtocol):
    """
    Driver implementing the ConsoleProtocol interface using a subprocess
    """
    cmd = attr.ib(validator=attr.validators.instance_of(str))
    txdelay = attr.ib(default=0.0, validator=attr.validators.instance_of(float))
    txchunk = attr.ib(default=1, validator=attr.validators.instance_of(int))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.status = 0
        self._child = None

    def open(self):
        """Starts the subprocess, does nothing if it is already open"""
        if self.status:
            return
        cmd = shlex.split(self.cmd)
        self._child = subprocess.Popen(
            cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=None, bufsize=0
        )

        # make stdout non-blocking
        stdout_fd = self._child.stdout.fileno()
        stdout_fl = fcntl.fcntl(stdout_fd, fcntl.F_GETFL) | os.O_NONBLOCK
        fcntl.fcntl(stdout_fd, fcntl.F_SETFL, stdout_fl)

        # prepare for timeout handing
        self._poll = select.poll()
        self._poll.register(self._child.stdout.fileno())
        self.status = 1

    def close(self):
        """Stops the subprocess, does nothing if it is already closed"""
        if not self.status:
            return
        if self._child.poll() is not None:
            self._child.communicate()
            raise ExecutionError("child has vanished")
        self._child.terminate()
        try:
            outs, errs = self._child.communicate(timeout=1)
        except subprocess.TimeoutExpired:
            self._child.kill()
            outs, errs = self._child.communicate()

        if outs:
            self.logger.info("child stdout while closing: %s", outs)
        if errs:
            self.logger.warning("child error while closing: %s", errs)

        self._child = None
        self.status = 0

    def _read(self, size: int = 1024, timeout: int = 0, max_size: int = None):
        """
        Reads 'size' bytes from the serialport

        Keyword Arguments:
        size -- amount of bytes to read, defaults to 1024
        max_size -- maximal amount of bytes to read
        """
        if max_size:
            read_size = min(size, max_size)
        else:
            read_size = size

        if self._child.poll() is not None:
            raise ExecutionError("child has vanished")
        if self._poll.poll(timeout):
            return self._child.stdout.read(read_size)

        return b''

    def _write(self, data: bytes):
        """
        Writes 'data' to the serialport

        Arguments:
        data -- data to write, must be bytes
        """
        if self._child.poll() is not None:
            raise ExecutionError("child has vanished")
        result = self._child.stdin.write(data)
        self._child.stdin.flush()
        return result

    def on_activate(self):
        self.open()

    def on_deactivate(self):
        self.close()



from labgrid.resource import RawSerialPort  # pylint: disable=import-error


class TestSerialPort:
    def test_instanziation(self):
        s = RawSerialPort(None, 'serial', 'port')
        assert (s.port == 'port')

    def test_instanziation_with(self, target):
        s = RawSerialPort(target, 'serial', 'port', 115200)
        assert (s.port == 'port')
        assert (s.speed == 115200)
        assert s in target.resources



import logging
import time

from labgrid import Target
from labgrid.logging import basicConfig, StepLogger
from labgrid.driver import GpioDigitalOutputDriver
from labgrid.resource import SysfsGPIO

# enable info logging
basicConfig(level=logging.INFO)

# show labgrid steps on the console
StepLogger.start()

t = Target("main")
r = SysfsGPIO(t, name=None, index=60)
d = GpioDigitalOutputDriver(t, name=None)

p = t.get_driver("DigitalOutputProtocol")
print(t.resources)
p.set(True)
print(p.get())
time.sleep(2)
p.set(False)
print(p.get())
time.sleep(2)
p.set(True)
print(p.get())



import pytest

from labgrid.resource import Resource, NetworkSerialPort, TFTPProvider
from labgrid.resource.remote import RemoteNetworkInterface, RemoteTFTPProvider
from labgrid.driver import Driver, SerialDriver, NetworkInterfaceDriver, TFTPProviderDriver
from labgrid.strategy import Strategy
from labgrid.binding import StateError


class ResourceA(Resource):
    pass


class DriverA(Driver):
    bindings = {"res": ResourceA}

    @Driver.check_bound
    def get_export_vars(self):
        return {
            "a": "b",
        }


class StrategyA(Strategy):
    bindings = {
        "drv": DriverA,
    }


def test_export(target):
    ra = ResourceA(target, "resource")
    d = DriverA(target, "driver")
    s = StrategyA(target, "strategy")

    exported = target.export()
    assert exported == {
        "LG__DRV_A": "b",
    }

    target.activate(d)
    with pytest.raises(StateError):
        d.get_export_vars()


class StrategyB(Strategy):
    bindings = {
        "drv": DriverA,
    }

    def prepare_export(self):
        return {
            self.drv: "custom_name",
        }


def test_export_custom(target):
    ra = ResourceA(target, "resource")
    d = DriverA(target, "driver")
    s = StrategyB(target, "strategy")

    exported = target.export()
    assert exported == {
        "LG__CUSTOM_NAME_A": "b",
    }


def test_export_network_serial(target):
    NetworkSerialPort(target, None, host='testhost', port=12345, speed=115200)
    SerialDriver(target, None)

    exported = target.export()
    assert exported == {
        'LG__SERIALDRIVER_HOST': 'testhost',
        'LG__SERIALDRIVER_PORT': '12345',
        'LG__SERIALDRIVER_PROTOCOL': 'rfc2217',
        'LG__SERIALDRIVER_SPEED': '115200'
    }


def test_export_remote_network_interface(target):
    RemoteNetworkInterface(target, None, host='testhost', ifname='wlan0')
    NetworkInterfaceDriver(target, "netif")

    exported = target.export()
    assert exported == {
        'LG__NETIF_HOST': 'testhost',
        'LG__NETIF_IFNAME': 'wlan0'
    }


def test_export_tftp_provider(target):
    TFTPProvider(target, None, internal='/srv/tftp/testboard/', external='testboard/')
    TFTPProviderDriver(target, "tftp")

    exported = target.export()
    assert exported == {
        'LG__TFTP_HOST': 'localhost',
        'LG__TFTP_INTERNAL': '/srv/tftp/testboard/',
        'LG__TFTP_EXTERNAL': 'testboard/',
    }


def test_export_remote_tftp_provider(target):
    RemoteTFTPProvider(target, None, host='testhost', internal='/srv/tftp/testboard/', external='testboard/')
    TFTPProviderDriver(target, "tftp")

    exported = target.export()
    assert exported == {
        'LG__TFTP_HOST': 'testhost',
        'LG__TFTP_INTERNAL': '/srv/tftp/testboard/',
        'LG__TFTP_EXTERNAL': 'testboard/',
    }



import pexpect


def test_client_help():
    with pexpect.spawn("python -m labgrid.remote.client --help") as spawn:
        spawn.expect("usage")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0
        assert spawn.signalstatus is None


def test_exporter_help():
    with pexpect.spawn("python -m labgrid.remote.exporter --help") as spawn:
        spawn.expect("usage")
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0
        assert spawn.signalstatus is None


def test_exporter_start_coordinator_unreachable(monkeypatch, tmpdir):
    monkeypatch.setenv("LG_COORDINATOR", "coordinator.invalid")

    config = "exports.yaml"
    p = tmpdir.join(config)
    p.write(
        """
    Testport:
        NetworkSerialPort:
          host: 'localhost'
          port: 4000
    """
    )

    with pexpect.spawn(f"python -m labgrid.remote.exporter {config}", cwd=tmpdir) as spawn:
        spawn.expect("coordinator is unavailable", timeout=10)
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 100, spawn.before


def test_exporter_coordinator_becomes_unreachable(coordinator, exporter):
    coordinator.suspend_tree()

    exporter.spawn.expect(pexpect.EOF, timeout=30)
    exporter.spawn.close()
    assert exporter.exitstatus == 100

    coordinator.resume_tree()



from .bareboxdriver import BareboxDriver
from .ubootdriver import UBootDriver
from .smallubootdriver import SmallUBootDriver
from .serialdriver import SerialDriver
from .shelldriver import ShellDriver
from .sshdriver import SSHDriver
from .externalconsoledriver import ExternalConsoleDriver
from .exception import CleanUpError, ExecutionError
from .fastbootdriver import AndroidFastbootDriver
from .dfudriver import DFUDriver
from .openocddriver import OpenOCDDriver
from .quartushpsdriver import QuartusHPSDriver
from .flashromdriver import FlashromDriver
from .onewiredriver import OneWirePIODriver
from .powerdriver import ManualPowerDriver, ExternalPowerDriver, \
                         DigitalOutputPowerDriver, YKUSHPowerDriver, \
                         USBPowerDriver, SiSPMPowerDriver, NetworkPowerDriver, \
                         PDUDaemonDriver
from .usbloader import MXSUSBDriver, IMXUSBDriver, BDIMXUSBDriver, RKUSBDriver, UUUDriver
from .usbsdmuxdriver import USBSDMuxDriver
from .usbsdwiredriver import USBSDWireDriver
from .usbsdwire3driver import USBSDWire3Driver
from .common import Driver
from .qemudriver import QEMUDriver
from .modbusdriver import ModbusCoilDriver, WaveShareModbusCoilDriver
from .modbusrtudriver import ModbusRTUDriver
from .sigrokdriver import SigrokDriver, SigrokPowerDriver, SigrokDmmDriver
from .usbstoragedriver import USBStorageDriver, Mode
from .resetdriver import DigitalOutputResetDriver
from .gpiodriver import GpioDigitalOutputDriver
from .filedigitaloutput import FileDigitalOutputDriver
from .serialdigitaloutput import SerialPortDigitalOutputDriver
from .xenadriver import XenaDriver
from .dockerdriver import DockerDriver
from .lxaiobusdriver import LXAIOBusPIODriver
from .lxausbmuxdriver import LXAUSBMuxDriver
from .pyvisadriver import PyVISADriver
from .usbhidrelay import HIDRelayDriver
from .flashscriptdriver import FlashScriptDriver
from .usbaudiodriver import USBAudioInputDriver
from .usbvideodriver import USBVideoDriver
from .httpvideodriver import HTTPVideoDriver
from .networkinterfacedriver import NetworkInterfaceDriver
from .provider import HTTPProviderDriver, NFSProviderDriver, TFTPProviderDriver
from .rawnetworkinterfacedriver import RawNetworkInterfaceDriver
from .mqtt import TasmotaPowerDriver
from .manualswitchdriver import ManualSwitchDriver
from .usbtmcdriver import USBTMCDriver
from .deditecrelaisdriver import DeditecRelaisDriver
from .dediprogflashdriver import DediprogFlashDriver
from .httpdigitaloutput import HttpDigitalOutputDriver
from .eth008digitaloutput import Eth008DigitalOutputDriver
from .laadriver import LAASerialDriver, LAAPowerDriver, \
                       LAAUSBGadgetMassStorageDriver, LAAUSBDriver, \
                       LAAButtonDriver, LAALedDriver, LAATempDriver, LAAWattDriver, \
                       LAAProviderDriver



"""The autoinstall.main module runs an installation script automatically on multiple targets."""

import ast
import argparse
import logging
import multiprocessing
import textwrap
from time import sleep

from .. import Environment, target_factory
from ..logging import basicConfig, StepLogger
from ..exceptions import NoResourceFoundError


class Handler(multiprocessing.Process):
    def __init__(self, env, args, name):
        super().__init__(name=name)
        self.env = env
        self.args = args
        self.config = env.config
        self.name = name

        self.context = {
            "name": self.name,
            "env": self.env,
            "config": self.config,
        }
        self.context.update(target_factory.resources)
        self.context.update(target_factory.drivers)

    def _get_function(self, name, context):
        snippet = self.config.data["autoinstall"].get(name)
        if not snippet:
            return None

        code = f"def {name}():\n{textwrap.indent(snippet, ' ')}"
        tree = ast.parse(code, filename=self.env.config_file)
        ast.increment_lineno(tree, snippet.start_mark.line)
        co = compile(tree, filename=self.env.config_file, mode="exec")

        stage = {}
        exec(co, context, stage)  # pylint: disable=exec-used
        return stage[name]

    def _get_setup_function(self):
        context = self.context
        context["log"] = self.log.getChild("setup")
        setup = self._get_function("setup", self.context)
        if setup is None:

            def setup():
                pass

        return setup

    def _get_handler_function(self):
        context = self.context.copy()
        context["log"] = self.log.getChild("handler")
        handler = self._get_function("handler", context)
        return handler

    def _get_initial_resource(self):
        cls = self.config.data["autoinstall"].get("initial-resource")
        if not cls:
            return None

        return self.target.get_resource(self.context[cls], wait_avail=False)

    def run(self):
        self.log = logging.getLogger(self.name)
        self.log.info("creating new handler (PID %s)", self.pid)

        try:
            self.target = self.env.get_target(self.name)
            self.context["target"] = self.target
            if self.target is None:
                raise KeyError
        except Exception:  # pylint: disable=broad-except
            self.log.exception("target creation failed")
            return

        self.setup = self._get_setup_function()
        self.setup()

        self.initial_resource = self._get_initial_resource()

        self.handler = self._get_handler_function()
        while self.run_once():
            sleep(3)

    def run_once(self):
        try:
            if self.initial_resource:
                self.log.info("waiting until %s is available", self.initial_resource.display_name)
                while True:
                    self.target.update_resources()
                    if self.initial_resource.avail:
                        break
                    sleep(0.25)

            self.log.info("starting handler")
            self.target.update_resources()
            result = self.handler()
            if result is not None:
                self.log.warning("unexpected return value from handler: %s", repr(result))
            self.log.info("completed handler")

            if self.initial_resource:
                self.log.info("waiting until %s is unavailable", self.initial_resource.display_name)
                while True:
                    self.target.update_resources()
                    if not self.initial_resource.avail:
                        break
                    sleep(0.25)

        except NoResourceFoundError as e:
            if e.filter and len(e.filter) > 1:
                self.log.warning("resources %s not found, restarting", e.filter)
            elif e.filter:
                self.log.warning("resource %s not found, restarting", next(iter(e.filter)))
            else:
                self.log.warning("resource not found, restarting")
        except Exception:  # pylint: disable=broad-except
            self.log.exception("handler failed")
            return False

        if self.args.once:
            self.log.info("stopping handler (--once)")
            return False

        return True


class Manager:
    def __init__(self, env, args):
        self.env = env
        self.args = args
        self.config = env.config
        self.log = logging.getLogger("manager")

    def configure(self):
        if not "autoinstall" in self.env.config.data:
            self.log.error("no 'autoinstall' section found in '%s'", self.env.config_file)
            return False

        if not "handler" in self.env.config.data["autoinstall"]:
            self.log.error("no 'handler' definition found in '%s'", self.env.config_file)
            return False

        self.handlers = {}
        for target_name in self.config.data.get("targets", {}).keys():
            self.handlers[target_name] = Handler(self.env, self.args, target_name)

        if not self.handlers:
            self.log.error("no targets found in '%s'", self.env.config_file)
            return False

        return True

    def start(self):
        for i, handler in enumerate(self.handlers.values()):
            if i:
                # give previous handler some time to start
                sleep(3)
            handler.daemon = True
            handler.start()

    def join(self):
        for handler in self.handlers.values():
            handler.join()


def main():
    basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--debug", action="store_true", default=False, help="enable debug mode")
    parser.add_argument("--once", action="store_true", default=False, help="handle each target only once")
    parser.add_argument("config", type=str, help="config file")

    args = parser.parse_args()

    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    env = Environment(config_file=args.config)

    StepLogger.start()

    manager = Manager(env, args)
    if not manager.configure():
        exit(1)
    manager.start()
    manager.join()


if __name__ == "__main__":
    main()



import logging
import shlex
from typing import Dict, Type, List
import attr

from ..binding import BindingMixin
from ..util.ssh import sshmanager


@attr.s(eq=False)
class Resource(BindingMixin):
    """
    Represents a resource which is used by drivers. It only stores information
    and does not implement any actual functionality.

    Resources can exist without a target, but they must be bound to one before
    use.

    Life cycle:

    - create
    - bind (n times)
    """
    avail = attr.ib(default=True, init=False, validator=attr.validators.instance_of(bool))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._parent = None

        logger_name = self.__class__.__name__
        if self.target:
            logger_name += f"({self.target.name})"
        if self.name:
            logger_name += f":{self.name}"
        self.logger = logging.getLogger(logger_name)

    @property
    def command_prefix(self):
        return []

    def wrap_command(self, command):
        """
        Returns the command as-is, no need for command wrapping for local resources.

        Must stay compatible with NetworkResource.wrap_command().
        """
        return command

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, value):
        assert self._parent is None
        self._parent = value

    def get_managed_parent(self):
        """
        For Resources which have been created at runtime, return the
        ManagedResource resource which created it.

        Returns None otherwise.
        """
        return self._parent.get_managed_parent() if self._parent else None

    def poll(self):
        managed_parent = self.get_managed_parent()
        if managed_parent:
            managed_parent.poll()

    def get_bound_resources(self):
        """
        Resources only return themselves in a set, drivers will combine those sets.
        """
        return {self}


@attr.s(eq=False)
class NetworkResource(Resource):
    """
    Represents a remote Resource available on another computer.

    This stores a command_prefix to describe how to connect to the remote
    computer.

    Args:
        host (str): remote host the resource is available on
    """
    host = attr.ib(validator=attr.validators.instance_of(str))

    @property
    def command_prefix(self):
        host = self.host

        if hasattr(self, 'extra'):
            if self.extra.get('proxy_required'):
                host = self.extra.get('proxy')

        conn = sshmanager.get(host)
        prefix = conn.get_prefix()

        return prefix + ['--']

    def wrap_command(self, command):
        """
        Returns shell-escaped command with prepended command_prefix.

        Must stay compatible with Resource.wrap_command().
        """
        shargs = [shlex.quote(a) for a in command]
        return self.command_prefix + shargs


@attr.s(eq=False)
class ResourceManager:
    instances: 'Dict[Type[ResourceManager], ResourceManager]' = {}

    @classmethod
    def get(cls) -> 'ResourceManager':
        instance = ResourceManager.instances.get(cls)
        if instance is None:
            instance = cls()
            ResourceManager.instances[cls] = instance
        return instance

    def __attrs_post_init__(self):
        self.resources: List[ManagedResource] = []
        self.logger = logging.getLogger(str(self))

    def _add_resource(self, resource: 'ManagedResource'):
        self.resources.append(resource)
        self.on_resource_added(resource)

    def on_resource_added(self, resource: 'ManagedResource'):
        pass

    def poll(self):
        pass


@attr.s(eq=False)
class ManagedResource(Resource):
    """
    Represents a resource which can appear and disappear at runtime. Every
    ManagedResource has a corresponding ResourceManager which handles these
    events.
    """
    manager_cls = ResourceManager

    timeout = attr.ib(default=2.0, init=False, validator=attr.validators.instance_of(float))

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.avail = False
        self.manager = self.manager_cls.get()
        self.manager._add_resource(self)

    def poll(self):
        self.manager.poll()

    def get_managed_parent(self):
        return self



"""Config convenience class

This class encapsulates access functions to the environment configuration

"""
import os
import warnings
from yaml import YAMLError
import attr

from .exceptions import NoConfigFoundError, InvalidConfigError
from .util.yaml import load, resolve_templates

@attr.s(eq=False)
class Config:
    filename = attr.ib(validator=attr.validators.instance_of(str))

    def __attrs_post_init__(self):
        # load and parse the yaml configuration file
        self.base = os.path.dirname(os.path.abspath(self.filename))
        try:
            with open(self.filename) as file:
                self.data = load(file)
        except FileNotFoundError:
            raise NoConfigFoundError(
                f"configuration file '{self.filename}' could not be found"
            )
        except YAMLError as err:
            raise InvalidConfigError(f"Error in configuration file: {err}")

        if self.data is None:
            raise InvalidConfigError("Configuration file is empty")

        substitutions = {
            'BASE': self.base,
        }
        # map LG_* variables from OS environment into YAML config file using !template $LG_*
        # Only map LG_*, to protect from weird things in environment
        for x in os.environ.keys():
            if x.startswith("LG_"):
                substitutions[x] = os.environ[x]

        try:
            resolve_templates(self.data, substitutions)
        except KeyError as e:
            raise InvalidConfigError(
                f"configuration file '{self.filename}' refers to unknown variable '{e.args[0]}'"
            )
        except ValueError as e:
            raise InvalidConfigError(
                f"configuration file '{self.filename}' is invalid: {e}"
            )

        if self.get_option("crossbar_url", ""):
            warnings.warn("Ignored option 'crossbar_url' in config, use 'coordinator_address' instead", UserWarning)

    def resolve_path(self, path):
        """Resolve an absolute path

        Args:
            path (str): path to resolve

        Returns:
            str: the absolute path
        """
        path = os.path.expanduser(path)
        path = os.path.expandvars(path)
        if os.path.isabs(path):
            return path

        return os.path.join(self.base, path)

    def resolve_path_str_or_list(self, path):
        """
        Resolves a single path or multiple paths. Always returns a list (containing a single or
        multiple resolved paths).

        Args:
            path (str, list): path(s) to resolve

        Returns:
            list: absolute path(s

        Raises:
            TypeError: if input is neither str nor list
        """
        if isinstance(path, str):
            return [self.resolve_path(path)]
        elif isinstance(path, list):
            return [self.resolve_path(p) for p in path]

        raise TypeError("path must be str or list")

    def get_tool(self, tool):
        """Retrieve an entry from the tools subkey

        Args:
            tool (str): the tool to retrieve the path for

        Returns:
            str: path to the requested tools
        """
        try:
            path = str(self.data['tools'][tool])
        except KeyError:
            return tool

        resolved = self.resolve_path(path)
        if os.path.exists(resolved):
            return resolved

        return path

    def get_image_path(self, kind):
        """Retrieve an entry from the images subkey

        Args:
            kind (str): the kind of the image to retrieve the path for

        Returns:
            str: path to the image

        Raises:
            KeyError: if the requested image can not be found in the
                configuration
        """
        try:
            path = str(self.data['images'][kind])
            return self.resolve_path(path)
        except KeyError as e:
            raise KeyError(f"no path configured for image '{kind}'") from e

    def get_path(self, kind):
        """Retrieve an entry from the paths subkey

        Args:
            kind (str): the type of path to retrieve the path for

        Returns:
            str: path to the path

        Raises:
            KeyError: if the requested image can not be found in the
                configuration
        """
        try:
            path = str(self.data['paths'][kind])
            return self.resolve_path(path)
        except KeyError as e:
            raise KeyError(f"no path configured for path '{kind}'") from e

    def get_option(self, name, default=None):
        """Retrieve an entry from the options subkey

        Args:
            name (str): name of the option
            default (any): A default parameter in case the option can not be
                found

        Returns:
            any: value of the option or default parameter

        Raises:
            KeyError: if the requested image can not be found in the
                configuration
        """
        try:
            return self.data['options'][name]
        except KeyError:
            if default is None:
                raise KeyError(f"no option '{name}' found in configuration")

            return default

    def set_option(self, name, value):
        """Set an entry in the options subkey

        Args:
            name (str): name of the option
            value (any): the new value
        """
        self.data.setdefault('options', {})[name] = value

    def get_target_option(self, target, name, default=None):
        """Retrieve an entry from the options subkey under the specified target
           subkey

        Args:
            target (str): name of the target
            name (str): name of the option
            default (any): A default parameter in case the option can not be
                found

        Returns:
            any: value of the option or default parameter

        Raises:
            KeyError: if the requested key can not be found in the
                configuration, or if the target can not be found in the
                configuration.
        """
        if target not in self.data['targets']:
            raise KeyError(f"No target '{target}' found in configuration")

        try:
            return self.data['targets'][target]['options'][name]
        except (KeyError, TypeError):
            # Empty target declarations become None in the target dict, hence
            # TypeError when we try to subscript it.
            if default is None:
                raise KeyError(f"No option '{name}' found in configuration for target '{target}'")
            else:
                return default

    def set_target_option(self, target, name, value):
        """Set an entry in the options subkey under the specified target subkey

        Args:
            target (str): name of the target
            name (str): name of the option
            value (any): the new value

        Raises:
            KeyError: if the requested target can not be found in the
                configuration
        """
        assert isinstance(target, str)
        assert isinstance(name, str)

        if target not in self.data['targets']:
            raise KeyError(f"No target '{target}' found in configuration")

        # Empty targets become None in the target dict. Delete it to enable
        # setdefault below to work on the actual default instead of None.
        if self.data['targets'][target] is None:
            del self.data['targets'][target]

        trg = self.data['targets'].setdefault(target, {})
        trg.setdefault('options', {})[name] = value

    def get_targets(self):
        return self.data.get('targets', {})

    def get_imports(self):
        """Helper function that returns the list of all imports

        Returns:
            List: List of files which should be imported
        """
        imports = []

        if isinstance(self.data.get('imports', []), str):
            raise KeyError("imports needs to be list not string")
        for user_import in self.data.get('imports', []):
            # Try to resolve the import to a .py file
            import_path = self.resolve_path(user_import)
            if import_path.endswith('.py'):
                imports.append(import_path)
            else:
                # Fallback to importing module if not a .py file path
                imports.append(user_import)

        return imports

    def get_paths(self):
        """Helper function that returns the subdict of all paths

        Returns:
            Dict: Dictionary containing all path definitions
        """
        paths = {}

        for name, path in self.data.get('paths', {}).items():
            paths[name] = self.resolve_path(path)

        return paths

    def get_images(self):
        """Helper function that returns the subdict of all images

        Returns:
            Dict: Dictionary containing all image definitions
        """
        images = {}

        for name, image in self.data.get('images', {}).items():
            images[name] = self.resolve_path(image)

        return images

    def get_features(self):
        return set(self.data.get('features', {}))



import attr

from ..factory import target_factory
from .common import Resource


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkService(Resource):
    address = attr.ib(validator=attr.validators.instance_of(str))
    username = attr.ib(validator=attr.validators.instance_of(str))
    password = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    port = attr.ib(default=22, validator=attr.validators.instance_of(int))



import attr


@attr.s(eq=False)
class NoConfigFoundError(Exception):
    msg = attr.ib(validator=attr.validators.instance_of(str))


@attr.s(eq=False)
class NoSupplierFoundError(Exception):
    msg = attr.ib(validator=attr.validators.instance_of(str))
    filter = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(set))
    )


@attr.s(eq=False)
class InvalidConfigError(Exception):
    msg = attr.ib(validator=attr.validators.instance_of(str))


@attr.s(eq=False)
class NoDriverFoundError(NoSupplierFoundError):
    pass


@attr.s(eq=False)
class NoResourceFoundError(NoSupplierFoundError):
    found = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(list))
    )


@attr.s(eq=False)
class NoStrategyFoundError(NoSupplierFoundError):
    pass


@attr.s(eq=False)
class RegistrationError(Exception):
    msg = attr.ib(validator=attr.validators.instance_of(str))



import pytest


@pytest.fixture(scope="function")
def bootloader(target, strategy, capsys):
    with capsys.disabled():
        strategy.transition("barebox")
    return target["CommandProtocol"]  # this will return the BareboxDriver


@pytest.fixture(scope="function")
def shell(target, strategy, capsys):
    with capsys.disabled():
        strategy.transition("shell")
    return target["CommandProtocol"]  # this will return the ShellDriver


def test_barebox(bootloader):
    stdout, stderr, returncode = bootloader.run("version")
    assert returncode == 0
    assert stdout
    assert not stderr
    assert "barebox" in "\n".join(stdout)


def test_shell(shell):
    stdout, stderr, returncode = shell.run("cat /proc/version")
    assert returncode == 0
    assert stdout
    assert not stderr
    assert "Linux" in stdout[0]


def test_barebox_2(bootloader):
    bootloader.run_check("true")



from importlib import import_module
import attr

from ..factory import target_factory
from ..protocol import DigitalOutputProtocol
from ..util.proxy import proxymanager
from .common import Driver
from .exception import ExecutionError


@target_factory.reg_driver
@attr.s(eq=False)
class ModbusCoilDriver(Driver, DigitalOutputProtocol):
    bindings = {
        "coil": "ModbusTCPCoil",
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._module = import_module("pyModbusTCP.client")
        self._consts = import_module("pyModbusTCP.constants")
        self.client = None

    def on_activate(self):
        # we can only forward if the backend knows which port to use
        host, port = proxymanager.get_host_and_port(self.coil, default_port=502)
        self.client = self._module.ModbusClient(host=host, port=int(port), auto_open=True, auto_close=True)

    def on_deactivate(self):
        self.client = None

    def _handle_error(self, action):
        error_code = self.client.last_error
        if error_code == self._consts.MB_EXCEPT_ERR:
            exc = self.client.last_except
            if exc not in [self._consts.EXP_ACKNOWLEDGE, self._consts.EXP_NONE]:
                raise ExecutionError(f"Could not {action} coil (code={error_code}/exception={exc})")
        raise ExecutionError(f"Could not {action} coil (code={error_code})")

    @Driver.check_active
    def set(self, status):
        write_status = None
        if self.coil.invert:
            status = not status
        if self.coil.write_multiple_coils:
            write_status = self.client.write_multiple_coils(self.coil.coil, [bool(status)])
        else:
            write_status = self.client.write_single_coil(self.coil.coil, bool(status))
        if write_status is None:
            self._handle_error("write")

    @Driver.check_active
    def get(self):
        status = self.client.read_coils(self.coil.coil)
        if status is None:
            self._handle_error("read")

        status = status[0]
        if self.coil.invert:
            status = not status
        return status


@target_factory.reg_driver
@attr.s(eq=False)
class WaveShareModbusCoilDriver(ModbusCoilDriver):
    """
    Waveshare Modbus Relay driver.

    Waveshare only implement the ability to query the status of all relays not just one.
    https://www.waveshare.com/wiki/Modbus_RTU_Relay#Read_States_of_Relays
    """

    bindings = {
        "coil": "WaveshareModbusTCPCoil",
    }

    @Driver.check_active
    def get(self):
        status = self.client.read_coils(0x0000, self.coil.coil_count)
        if status is None:
            self._handle_error("read")

        status = status[self.coil.coil]
        if self.coil.invert:
            status = not status
        return status

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self.coil.coil >= self.coil.coil_count:
            raise ValueError("Coil exceeds coil count")



"""
Auxiliary classes that assist DockerDriver. Specifically, DockerDaemon
and DockerManager will create the NetworkResource instance that is declared
in the specification (e.g. yaml) of DockerDriver.
"""

import socket

import attr

from labgrid.util.dict import find_dict
from ..factory import target_factory
from .common import ManagedResource, ResourceManager


class DockerConstants:
    """Class constants for handling container cleanup"""

    DOCKER_LG_CLEANUP_LABEL = "lg_cleanup"
    # Currently only a single cleanup routine is implemented which removes
    # all labgrid created containers on each test run. In the future more
    # sophisticated mechanisms could be implemented by adding
    # more cleanup types.
    DOCKER_LG_CLEANUP_TYPE_AUTO = "auto"


@attr.s(eq=False)
class DockerManager(ResourceManager):
    """The DockerManager is responsible for cleaning up dangling
    containers and managing the managed NetworkService
    resources. Different docker daemons for different targets are
    allowed, but there has to be only one for each target.
    """

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._client = dict()
        self._docker_daemons_cleaned = list()

    def on_resource_added(self, resource):
        """If the resource added is a DockerDaemon, make sure this is the
        only one added for this target.  If it is, create a docker
        client to be used to communicate with the docker daemon.

        """
        import docker  # lazy import!

        if resource.target.name in self._client:
            raise EnvironmentError("Only one docker daemon is allowed for each target")
        self._client[resource.target.name] = docker.DockerClient(base_url=resource.docker_daemon_url)
        self._container_cleanup(self._client[resource.target.name])
        resource.avail = True

    def poll(self):
        """Ask associated DockerDaemon resource to check if associated
        NetworkService has come up."""
        for resource in self.resources:
            resource.on_poll(self._client[resource.target.name])

    def _container_cleanup(self, docker_client):
        """Fetches a list of containers from the daemon. Each container that
        has the label DOCKER_LG_CLEANUP_LABEL is removed. This is for
        cleanup of potential dangling containers left over from
        earlier test runs.
        """
        if docker_client.api.base_url not in self._docker_daemons_cleaned:
            container_list = docker_client.api.containers(
                all=True, filters={"label": DockerConstants.DOCKER_LG_CLEANUP_LABEL}
            )
            for container in container_list:
                if (
                    container["Labels"][DockerConstants.DOCKER_LG_CLEANUP_LABEL]
                    == DockerConstants.DOCKER_LG_CLEANUP_TYPE_AUTO
                ):
                    self.logger.info("Deleting container %s", container["Names"][0])
                    docker_client.api.remove_container(container["Id"], force=True)
            self._docker_daemons_cleaned.append(docker_client.api.base_url)


@target_factory.reg_resource
@attr.s(eq=False)
class DockerDaemon(ManagedResource):
    """A resource identifying a docker daemon"""

    docker_daemon_url = attr.ib(validator=attr.validators.instance_of(str))

    """The docker network service is a managed resource mapping a
    container name to a network service.
    """
    manager_cls = DockerManager

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._nw_services = dict()
        self.timeout = 5.0
        self.avail = True

    def on_client_bound(self, client):
        """Each time a docker driver binds to the DockerDaemon resource
        the docker driver network services list is iterated and for each
        network service defined a NetworkService resource instance is
        created with the parameters from the configuration file.

        """
        for network_service in client.network_services:
            # IP address is not yet known at this point
            network_service["address"] = ""
            if "name" in network_service:
                service_name = network_service["name"]
            else:
                service_name = "NetworkService"
            nw_service = target_factory.make_resource(self.target, "NetworkService", service_name, network_service)
            nw_service.parent = self
            if client.container_name not in self._nw_services:
                self._nw_services[client.container_name] = list()
            self._nw_services[client.container_name].append(nw_service)

    def on_poll(self, docker_client):
        """Check if associated NetworkService has come up."""
        self._update_resource(docker_client)

    def _update_resource(self, docker_client):
        """Update resource takes a docker client and uses this to lookup a
        container by name. If the container exists the IP address is
        looked up by inspecting the docker container. If an IP address
        is found the address is filled in the NetworkService base
        class and the resource is identified as available.
        :param docker_client: The docker client to use for lookup
        """
        for container_name, nw_service_list in self._nw_services.items():
            for nw_service in nw_service_list:
                if nw_service.address == "":
                    container = docker_client.api.containers(filters={"name": "/" + container_name})
                    self.logger.debug("Containers found %s", container)
                    if container:
                        nw_service.address = find_dict(d=container[0]["NetworkSettings"], key="IPAddress")
                if nw_service.address != "" and self._socket_connect(nw_service.address, nw_service.port):
                    nw_service.avail = True
                else:
                    nw_service.avail = False

    @staticmethod
    def _socket_connect(address, port):
        """
        Try to do a socket connect
        :param address: The ip address to connect to
        :param port: The port to connect to
        :return: True on successful connection, False if failed
        """
        try:
            s = socket.create_connection((address, port))
        except Exception:
            return False
        s.shutdown(socket.SHUT_RDWR)
        s.close()
        return True



import contextlib

import attr

from .common import Driver
from .exception import ExecutionError
from ..factory import target_factory
from ..step import step
from ..util.agentwrapper import AgentWrapper
from ..util.ssh import sshmanager
from ..resource.remote import RemoteNetworkInterface


@target_factory.reg_driver
@attr.s(eq=False)
class NetworkInterfaceDriver(Driver):
    bindings = {
        "iface": {"RemoteNetworkInterface", "NetworkInterface", "USBNetworkInterface"},
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.wrapper = None

    def on_activate(self):
        if isinstance(self.iface, RemoteNetworkInterface):
            host = self.iface.host
            self.ssh = sshmanager.get(host)
        else:
            host = None
            # port forwarding is not useful for localhost
            self.ssh = None
        self.wrapper = AgentWrapper(host)
        self.proxy = self.wrapper.load("network_interface")

    def on_deactivate(self):
        try:
            self.proxy.disable(self.iface.ifname)
        finally:
            self.wrapper.close()
            self.wrapper = None
            self.proxy = None

    @property
    def skip_deactivate_on_export(self):
        # We need to keep the agent and SSH exports active.
        return True

    def get_export_vars(self):
        return {
            "host": self.iface.host,
            "ifname": self.iface.ifname or "",
        }

    # basic
    @Driver.check_active
    @step()
    def configure(self, settings):
        "Set a configuration on this interface and activate it."
        self.proxy.configure(self.iface.ifname, settings)

    @Driver.check_active
    @step(args=["expected"])
    def wait_state(self, expected, timeout=60):
        """Wait until the expected state is reached or the timeout expires.

        Possible states include disconnected and activated. See NM's
        DeviceState for more details.
        """
        self.proxy.wait_state(self.iface.ifname, expected, timeout)

    @Driver.check_active
    @step()
    def disable(self):
        "Disable and remove the created labgrid connection."
        self.proxy.disable(self.iface.ifname)

    @Driver.check_active
    @step()
    def get_active_settings(self):
        "Get the currently active settings from this interface."
        return self.proxy.get_active_settings(self.iface.ifname)

    @Driver.check_active
    @step()
    def get_settings(self):
        "Get the settings of the labgrid connection for this interface."
        return self.proxy.get_settings(self.iface.ifname)

    @Driver.check_active
    @step()
    def get_state(self):
        """Get the current state of this interface.

        This includes information such as the addresses or WiFi connection.
        """
        return self.proxy.get_state(self.iface.ifname)

    # dhcpd
    @Driver.check_active
    @step()
    def get_dhcpd_leases(self):
        """Get the list of active DHCP leases in shared mode.

        This requires read access to /var/lib/NetworkManager/dnsmasq-* on the
        exporter.
        """

        return self.proxy.get_dhcpd_leases(self.iface.ifname)

    # wireless
    @Driver.check_active
    @step()
    def request_scan(self):
        "Request a scan to update the list visible access points."
        return self.proxy.request_scan(self.iface.ifname)

    @Driver.check_active
    @step()
    def get_access_points(self, scan=None):
        """Get the list of currently visible access points.

        When scan is None (the default), it will automatically trigger a scan
        if the last scan is more than 30 seconds old.
        """
        return self.proxy.get_access_points(self.iface.ifname, scan)

    @Driver.check_active
    @contextlib.contextmanager
    def forward_local(self, remote_host, remote_port, local_port=None):
        """Forward a local port to an address and port via the exporter.

        A context manager that keeps the local port forwarded as long as the
        context remains valid. A connection can be made to the returned port on
        localhost and it will be forwarded to the address and port.

        If localport is not set, a random free port will be selected.

        usage:
            with netif.forward_local('127.0.0.1', 8080, local_port=1234) as local_port:
                # Use localhost:1234 here to connect to port 8080 on
                # the exporter

            with netif.forward_local('192.168.1.2', 8080) as local_port:
                # Use localhost:local_port here to connect to port 8080 on
                # 192.168.1.2

        returns:
            local_port
        """
        if not self.ssh:
            raise ExecutionError("Resource is not on an exporter")

        local_port = self.ssh.add_port_forward(remote_host, remote_port, local_port)
        try:
            yield local_port
        finally:
            self.ssh.remove_port_forward(remote_host, remote_port)

    @Driver.check_active
    @contextlib.contextmanager
    def forward_remote(self, remote_port, local_port, remote_bind=None):
        """Forward a remote port on the exporter to a local port.

        A context manager that keeps a remote port forwarded to a local port as
        long as the context remains valid. A connection can be made to the
        remote port on the target device will be forwarded to the returned local
        port on localhost.

        Note that the remote socket is not *bound* to any specific IP by
        default, making it reachable by the target. Also, 'GatewayPorts
        clientspecified' needs to be configured in the remote host's
        sshd_config.

        usage:
            with netif.forward_remote(8080, 8081) as local_port:
                # Connections to port 8080 on the exporter will be redirected to
                # localhost:8081
        """
        if not self.ssh:
            raise ExecutionError("Resource is not on an exporter")

        self.ssh.add_remote_port_forward(remote_port, local_port, remote_bind)
        try:
            yield
        finally:
            self.ssh.remove_remote_port_forward(remote_port, local_port, remote_bind)



from ..exception import ExecutionError
from ...util.snmp import SimpleSNMP


OID = ".1.3.6.1.4.1.534.6.6.7.6.6.1"
NUMBER_OF_OUTLETS = 16


def power_set(host, port, index, value):
    assert 1 <= int(index) <= NUMBER_OF_OUTLETS

    _snmp = SimpleSNMP(host, 'public', port=port)
    cmd_id = 4 if int(value) else 3
    outlet_control_oid = "{}.{}.0.{}".format(OID, cmd_id, index)

    _snmp.set(outlet_control_oid, 1)


def power_get(host, port, index):
    assert 1 <= int(index) <= NUMBER_OF_OUTLETS

    _snmp = SimpleSNMP(host, 'public', port=port)
    output_status_oid = "{}.2.0.{}".format(OID, index)

    value = _snmp.get(output_status_oid)

    if value == 1:  # On
        return True
    if value == 0:  # Off
        return False

    if value == 3:  # Pending on - treat as on
        return True
    if value == 2:  # Pending off - treat as off
        return False

    raise ExecutionError("failed to get SNMP value")



import os

from urllib.parse import urlsplit, urlunsplit, urlparse

from .ssh import sshmanager

from ..resource.common import Resource

__all__ = ["proxymanager"]


class ProxyError(Exception):
    pass


class ProxyManager:
    """The ProxyManager class is only used inside labgrid.util.proxy (similar
    to a singleton), don't instanciate this class, use the exported
    proxymanager instead."""

    _force_proxy = os.environ.get("LG_PROXY", None)

    @classmethod
    def force_proxy(cls, force_proxy):
        assert isinstance(force_proxy, str)
        cls._force_proxy = force_proxy

    @classmethod
    def get_host_and_port(cls, res, *, default_port=None, force_port=None):
        """get host and port for a proxy connection from a Resource

        Args:
            res (Resource): The resource to retrieve the proxy for
            default_port (optional): TCP port to use if no other port is
                configured
            force_port (optional): TCP port to use instead of the one
                configured for the resource

        Returns:
            (host, port) host and port for the proxy connection

        Raises:
            ExecutionError: if the SSH connection/forwarding fails
        """
        assert isinstance(res, Resource)

        prefix = "" if "//" in res.host else "//"
        s = urlparse(prefix + res.host)
        host = s.hostname
        if force_port:
            port = force_port
        elif s.port:
            port = s.port
        else:
            port = getattr(res, "port", None) or default_port

        extra = getattr(res, "extra", {})
        if extra:
            proxy_required = extra.get("proxy_required")
            proxy = extra.get("proxy")
            if proxy_required:
                port = sshmanager.request_forward(proxy, host, port)
                host = "localhost"
                return host, port

        if cls._force_proxy:
            port = sshmanager.request_forward(cls._force_proxy, host, port)
            host = "localhost"

        return host, port

    @classmethod
    def get_url(cls, url, *, default_port=None):
        assert isinstance(url, str)

        s = urlsplit(url)

        hostname = s.hostname
        port = s.port or default_port

        if hostname is None:
            raise ProxyError(f"Invalid url: {url} does not contain a host")

        if port is None:
            raise ProxyError(f"Invalid url: {url} does not contain a port and no default set")

        if cls._force_proxy:
            port = sshmanager.request_forward(cls._force_proxy, hostname, port)
            hostname = "localhost"

        if ":" in hostname:
            # IPv6 address
            s = s._replace(netloc=f"[{hostname}]:{port}")
        else:
            s = s._replace(netloc=f"{hostname}:{port}")

        return urlunsplit(s)

    @classmethod
    def get_grpc_address(cls, address, *, default_port=None):
        url = f"//{address}"
        url = proxymanager.get_url(url, default_port=default_port)
        address = url.lstrip("/")
        return address

    @classmethod
    def get_command(cls, res, host, port, ifname=None):
        """get argument list to start a proxy process connected to the target"""
        assert isinstance(res, Resource)

        proxy = cls._force_proxy

        extra = getattr(res, "extra", {})
        if extra.get("proxy_required") or proxy:  # use specific proxy when needed
            proxy = extra.get("proxy") or proxy

        if not proxy:
            return None

        conn = sshmanager.get(proxy)
        command = conn.get_prefix()
        if ifname:
            command += [
                "--",
                "sudo",
                "--non-interactive",
                "labgrid-bound-connect",
                ifname,
                host,
                str(port),
            ]
        else:
            if ":" in host:  # IPv6
                host = f"[{host}]"
            command += ["-W", f"{host}:{port}"]
        return command


proxymanager = ProxyManager()



import asyncio
import time
import enum
import random
import re
import string
import logging
from datetime import datetime
from fnmatch import fnmatchcase

import attr

from .generated import labgrid_coordinator_pb2

__all__ = [
    "TAG_KEY",
    "TAG_VAL",
    "ResourceEntry",
    "ResourceMatch",
    "Place",
    "ReservationState",
    "Reservation",
]

TAG_KEY = re.compile(r"[a-z][a-z0-9_]+")
TAG_VAL = re.compile(r"[a-z0-9_]?")


def set_map_from_dict(m, d):
    for k, v in d.items():
        assert isinstance(k, str)
        if v is None:
            m[k].Clear()
        elif isinstance(v, bool):
            m[k].bool_value = v
        elif isinstance(v, int):
            if v < 0:
                m[k].int_value = v
            else:
                m[k].uint_value = v
        elif isinstance(v, float):
            m[k].float_value = v
        elif isinstance(v, str):
            m[k].string_value = v
        else:
            raise ValueError(f"cannot translate {repr(v)} to MapValue")


def build_dict_from_map(m):
    d = {}
    for k, v in m.items():
        v: labgrid_coordinator_pb2.MapValue
        kind = v.WhichOneof("kind")
        if kind is None:
            d[k] = None
        else:
            d[k] = getattr(v, kind)
    return d


@attr.s(eq=False)
class ResourceEntry:
    data = attr.ib()  # cls, params

    def __attrs_post_init__(self):
        assert isinstance(self.data, dict)
        self.data.setdefault("acquired", None)
        self.data.setdefault("avail", False)

    @property
    def acquired(self):
        return self.data["acquired"]

    @property
    def avail(self):
        return self.data["avail"]

    @property
    def cls(self):
        return self.data["cls"]

    @property
    def params(self):
        return self.data["params"]

    @property
    def args(self):
        """arguments for resource construction"""
        args = self.data["params"].copy()
        args.pop("extra", None)
        return args

    @property
    def extra(self):
        """extra resource information"""
        return self.data["params"].get("extra", {})

    def asdict(self):
        return {
            "cls": self.cls,
            "params": self.params,
            "acquired": self.acquired,
            "avail": self.avail,
        }

    def update(self, data):
        """apply updated information from the exporter on the coordinator"""
        data = data.copy()
        data.setdefault("acquired", None)
        data.setdefault("avail", False)
        self.data = data

    def acquire(self, place_name):
        assert self.data["acquired"] is None
        self.data["acquired"] = place_name

    def release(self):
        # ignore repeated releases
        self.data["acquired"] = None

    def as_pb2(self):
        msg = labgrid_coordinator_pb2.Resource()
        msg.cls = self.cls
        params = self.params.copy()
        extra = params.pop("extra", {})
        set_map_from_dict(msg.params, params)
        set_map_from_dict(msg.extra, extra)
        if self.acquired is not None:
            msg.acquired = self.acquired
        msg.avail = self.avail
        return msg

    @staticmethod
    def data_from_pb2(pb2):
        assert isinstance(pb2, labgrid_coordinator_pb2.Resource)
        data = {
            "cls": pb2.cls,
            "params": build_dict_from_map(pb2.params),
            "acquired": pb2.acquired or None,
            "avail": pb2.avail,
        }
        data["params"]["extra"] = build_dict_from_map(pb2.extra)
        return data

    @classmethod
    def from_pb2(cls, pb2):
        assert isinstance(pb2, labgrid_coordinator_pb2.Place)
        return cls(cls.data_from_pb2(pb2))


@attr.s(eq=True, repr=False, str=False)
# This class requires eq=True, since we put the matches into a list and require
# the cmp functions to be able to remove the matches from the list later on.
class ResourceMatch:
    exporter = attr.ib()
    group = attr.ib()
    cls = attr.ib()
    name = attr.ib(default=None, order=False)
    # rename is just metadata, so don't use it for comparing matches
    rename = attr.ib(default=None, eq=False)

    @classmethod
    def fromstr(cls, pattern):
        if not 2 <= pattern.count("/") <= 3:
            raise ValueError(f"invalid pattern format '{pattern}' (use 'exporter/group/cls/name')")
        return cls(*pattern.split("/"))

    def __repr__(self):
        result = f"{self.exporter}/{self.group}/{self.cls}"
        if self.name is not None:
            result += f"/{self.name}"
        return result

    def __str__(self):
        result = repr(self)
        if self.rename:
            result += " -> " + self.rename
        return result

    def ismatch(self, resource_path):
        """Return True if this matches the given resource"""
        try:
            exporter, group, cls, name = resource_path
        except ValueError:
            exporter, group, cls = resource_path
            name = None

        if not fnmatchcase(exporter, self.exporter):
            return False
        if not fnmatchcase(group, self.group):
            return False
        if not fnmatchcase(cls, self.cls):
            return False
        if name and self.name and not fnmatchcase(name, self.name):
            return False

        return True

    def as_pb2(self):
        return labgrid_coordinator_pb2.ResourceMatch(
            exporter=self.exporter,
            group=self.group,
            cls=self.cls,
            name=self.name,
            rename=self.rename,
        )

    @classmethod
    def from_pb2(cls, pb2):
        assert isinstance(pb2, labgrid_coordinator_pb2.ResourceMatch)
        return cls(
            exporter=pb2.exporter,
            group=pb2.group,
            cls=pb2.cls,
            name=pb2.name if pb2.HasField("name") else None,
            rename=pb2.rename,
        )


@attr.s(eq=False)
class Place:
    name = attr.ib()
    aliases = attr.ib(default=attr.Factory(set), converter=set)
    comment = attr.ib(default="")
    tags = attr.ib(default=attr.Factory(dict))
    matches = attr.ib(default=attr.Factory(list))
    acquired = attr.ib(default=None)
    acquired_resources = attr.ib(default=attr.Factory(list))
    allowed = attr.ib(default=attr.Factory(set), converter=set)
    created = attr.ib(default=attr.Factory(time.time))
    changed = attr.ib(default=attr.Factory(time.time))
    reservation = attr.ib(default=None)

    def asdict(self):
        # in the coordinator, we have resource objects, otherwise just a path
        acquired_resources = []
        for resource in self.acquired_resources:
            if isinstance(resource, (tuple, list)):
                acquired_resources.append(resource)
            else:
                acquired_resources.append(resource.path)

        return {
            "aliases": list(self.aliases),
            "comment": self.comment,
            "tags": self.tags,
            "matches": [attr.asdict(x) for x in self.matches],
            "acquired": self.acquired,
            "acquired_resources": acquired_resources,
            "allowed": list(self.allowed),
            "created": self.created,
            "changed": self.changed,
            "reservation": self.reservation,
        }

    def update_from_pb2(self, place_pb2):
        # FIXME untangle this...
        place = Place.from_pb2(place_pb2)
        fields = attr.fields_dict(type(self))
        for k, v in place.asdict().items():
            assert k in fields
            if k == "name":
                # we cannot rename places
                assert v == self.name
                continue
            if k == "matches":
                self.matches = [ResourceMatch.from_pb2(m) for m in place_pb2.matches]
                continue
            setattr(self, k, v)

    def show(self, level=0):
        indent = "  " * level
        if self.aliases:
            print(indent + f"aliases: {', '.join(sorted(self.aliases))}")
        if self.comment:
            print(indent + f"comment: {self.comment}")
        if self.tags:
            print(indent + f"tags: {', '.join(k + '=' + v for k, v in sorted(self.tags.items()))}")
        print(indent + "matches:")
        for match in sorted(self.matches):
            print(indent + f"  {match}")
        print(indent + f"acquired: {self.acquired}")
        print(indent + "acquired resources:")
        # in the coordinator, we have resource objects, otherwise just a path
        for resource in sorted(self.acquired_resources):
            if isinstance(resource, (tuple, list)):
                resource_path = resource
            else:
                resource_path = resource.path
            match = self.getmatch(resource_path)
            if match.rename:
                print(indent + f"  {'/'.join(resource_path)} -> {match.rename}")
            else:
                print(indent + f"  {'/'.join(resource_path)}")
        if self.allowed:
            print(indent + f"allowed: {', '.join(self.allowed)}")
        print(indent + f"created: {datetime.fromtimestamp(self.created)}")
        print(indent + f"changed: {datetime.fromtimestamp(self.changed)}")
        if self.reservation:
            print(indent + f"reservation: {self.reservation}")

    def getmatch(self, resource_path):
        """Return the ResourceMatch object for the given resource path or None if not found.

        A resource_path has the structure (exporter, group, cls, name).
        """
        for match in self.matches:
            if match.ismatch(resource_path):
                return match

        return None

    def hasmatch(self, resource_path):
        """Return True if this place as a ResourceMatch object for the given resource path.

        A resource_path has the structure (exporter, group, cls, name).
        """
        return self.getmatch(resource_path) is not None

    def unmatched(self, resource_paths):
        """Returns a match which could not be matched to the list of resource_path

        A resource_path has the structure (exporter, group, cls, name).
        """
        for match in self.matches:
            if not any([match.ismatch(resource) for resource in resource_paths]):
                return match

    def touch(self):
        self.changed = time.time()

    def as_pb2(self):
        try:
            acquired_resources = []
            for resource in self.acquired_resources:
                assert not isinstance(resource, (tuple, list)), "as_pb2() only implemented for coordinator"
                assert len(resource.path) == 4
                path = "/".join(resource.path)
                acquired_resources.append(path)

            place = labgrid_coordinator_pb2.Place()
            place.name = self.name
            place.aliases.extend(self.aliases)
            place.comment = self.comment
            place.matches.extend(m.as_pb2() for m in self.matches)
            place.acquired = self.acquired or ""
            place.acquired_resources.extend(acquired_resources)
            place.allowed.extend(self.allowed)
            place.changed = self.changed
            place.created = self.created
            if self.reservation:
                place.reservation = self.reservation
            for key, value in self.tags.items():
                place.tags[key] = value
            return place
        except TypeError:
            logging.exception("failed to convert place %s to protobuf", self)
            raise

    @classmethod
    def from_pb2(cls, pb2):
        assert isinstance(pb2, labgrid_coordinator_pb2.Place)
        acquired_resources = []
        for path in pb2.acquired_resources:
            path = path.split("/")
            assert len(path) == 4
            acquired_resources.append(path)
        return cls(
            name=pb2.name,
            aliases=pb2.aliases,
            comment=pb2.comment,
            tags=dict(pb2.tags),
            matches=[ResourceMatch.from_pb2(m) for m in pb2.matches],
            acquired=pb2.acquired if pb2.HasField("acquired") and pb2.acquired else None,
            acquired_resources=acquired_resources,
            allowed=pb2.allowed,
            created=pb2.created,
            changed=pb2.changed,
            reservation=pb2.reservation if pb2.HasField("reservation") else None,
        )


class ReservationState(enum.Enum):
    waiting = 0
    allocated = 1
    acquired = 2
    expired = 3
    invalid = 4


@attr.s(eq=False)
class Reservation:
    owner = attr.ib(validator=attr.validators.instance_of(str))
    token = attr.ib(
        default=attr.Factory(lambda: "".join(random.choice(string.ascii_uppercase + string.digits) for i in range(10)))
    )
    state = attr.ib(
        default="waiting",
        converter=lambda x: x if isinstance(x, ReservationState) else ReservationState[x],
        validator=attr.validators.instance_of(ReservationState),
    )
    prio = attr.ib(default=0.0, validator=attr.validators.instance_of(float))
    # a dictionary of name -> filter dicts
    filters = attr.ib(default=attr.Factory(dict), validator=attr.validators.instance_of(dict))
    # a dictionary of name -> place names
    allocations = attr.ib(default=attr.Factory(dict), validator=attr.validators.instance_of(dict))
    created = attr.ib(default=attr.Factory(time.time))
    timeout = attr.ib(default=attr.Factory(lambda: time.time() + 60))

    def asdict(self):
        return {
            "owner": self.owner,
            "state": self.state.name,
            "prio": self.prio,
            "filters": self.filters,
            "allocations": self.allocations,
            "created": self.created,
            "timeout": self.timeout,
        }

    def refresh(self, delta=60):
        self.timeout = max(self.timeout, time.time() + delta)

    @property
    def expired(self):
        return self.timeout < time.time()

    def show(self, level=0):
        indent = "  " * level
        print(indent + f"owner: {self.owner}")
        print(indent + f"token: {self.token}")
        print(indent + f"state: {self.state.name}")
        if self.prio:
            print(indent + f"prio: {self.prio}")
        print(indent + "filters:")
        for name, fltr in self.filters.items():
            print(indent + f"  {name}: {' '.join([(k + '=' + v) for k, v in fltr.items()])}")
        if self.allocations:
            print(indent + "allocations:")
            for name, allocation in self.allocations.items():
                print(indent + f"  {name}: {', '.join(allocation)}")
        print(indent + f"created: {datetime.fromtimestamp(self.created)}")
        print(indent + f"timeout: {datetime.fromtimestamp(self.timeout)}")

    def as_pb2(self):
        res = labgrid_coordinator_pb2.Reservation()
        res.owner = self.owner
        res.token = self.token
        res.state = self.state.value
        res.prio = self.prio
        for name, fltr in self.filters.items():
            res.filters[name].CopyFrom(labgrid_coordinator_pb2.Reservation.Filter(filter=fltr))
        if self.allocations:
            # TODO: refactor to have only one place per filter group
            assert len(self.allocations) == 1
            assert "main" in self.allocations
            allocation = self.allocations["main"]
            assert len(allocation) == 1
            res.allocations.update({"main": allocation[0]})
        res.created = self.created
        res.timeout = self.timeout
        return res

    @classmethod
    def from_pb2(cls, pb2: labgrid_coordinator_pb2.Reservation):
        filters = {}
        for name, fltr_pb2 in pb2.filters.items():
            filters[name] = dict(fltr_pb2.filter)
        allocations = {}
        for fltr_name, place_name in pb2.allocations.items():
            allocations[fltr_name] = [place_name]
        return cls(
            owner=pb2.owner,
            token=pb2.token,
            state=ReservationState(pb2.state),
            prio=pb2.prio,
            filters=filters,
            allocations=allocations,
            created=pb2.created,
            timeout=pb2.timeout,
        )


async def queue_as_aiter(q):
    try:
        while True:
            try:
                item = await q.get()
            except asyncio.CancelledError:
                # gRPC doesn't like to receive exceptions from the request_iterator
                return
            if item is None:
                return
            yield item
            q.task_done()
            logging.debug("sent message %s", item)
    except Exception:
        logging.exception("error in queue_as_aiter")
        raise



from time import monotonic
from importlib import import_module

import attr

from ..factory import target_factory
from .common import ManagedResource, ResourceManager


@attr.s(eq=False)
class LXAIOBusNodeManager(ResourceManager):
    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._requests = import_module("requests")

        self._last = 0.0

    def _get_nodes(self, host):
        try:
            r = self._requests.get(f"http://{host}/nodes/")
            r.raise_for_status()
            j = r.json()
            return j["result"]
        except self._requests.exceptions.ConnectionError:
            self.logger.exception("failed to connect to host %s", host)
            return []

    def poll(self):
        if monotonic() - self._last < 2:
            return  # ratelimit requests
        self._last = monotonic()
        hosts = {r.host for r in self.resources}
        nodes = {h: self._get_nodes(h) for h in hosts}
        for resource in self.resources:
            resource.avail = resource.node in nodes[resource.host]


@attr.s(eq=False)
class LXAIOBusNode(ManagedResource):
    """This resource describes a generic LXA IO BUs Node.

    Args:
        host (str): hostname of the owserver e.g. localhost:4304
        node (str): node name e.g. EthMux-5c12ca8b"""

    manager_cls = LXAIOBusNodeManager

    host = attr.ib(validator=attr.validators.instance_of(str))
    node = attr.ib(validator=attr.validators.instance_of(str))

    def __attrs_post_init__(self):
        self.timeout = 30.0
        super().__attrs_post_init__()


@target_factory.reg_resource
@attr.s(eq=False)
class LXAIOBusPIO(LXAIOBusNode):
    """This resource describes a LXA IO Bus PIO Port.

    Args:
        pin (str): pin label e.g. OUT0
        invert (bool): optional, whether the logic level is inverted (active-low)"""

    pin = attr.ib(validator=attr.validators.instance_of(str))
    invert = attr.ib(default=False, validator=attr.validators.instance_of(bool))



import re
from importlib import import_module

import attr

from ..factory import target_factory
from ..protocol import DigitalOutputProtocol
from ..step import step
from ..util.proxy import proxymanager
from .common import Driver
from .exception import ExecutionError


@target_factory.reg_driver
@attr.s(eq=False)
class HttpDigitalOutputDriver(Driver, DigitalOutputProtocol):
    bindings = {"http": "HttpDigitalOutput"}

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._requests = import_module("requests")

    def on_activate(self):
        self._url_set = proxymanager.get_url(
            self.http.url,
            default_port=(443 if self.http.url.startswith("https") else 80),
        )

        if self.http.url_get:
            self._url_get = proxymanager.get_url(
                self.http.url_get,
                default_port=(443 if self.http.url_get.startswith("https") else 80),
            )

        else:
            self._url_get = self._url_set

    @Driver.check_active
    @step(args=["status"])
    def set(self, status):
        method = self.http.method or "PUT"
        body = self.http.body_asserted if status else self.http.body_deasserted

        res = self._requests.request(method, self._url_set, data=body)
        res.raise_for_status()

    @Driver.check_active
    @step(result=["True"])
    def get(self):
        res = self._requests.get(self._url_get)
        res.raise_for_status()

        # Check if the response body matches an asserted state
        if self.http.body_get_asserted:
            if re.fullmatch(self.http.body_get_asserted, res.text) is not None:
                return True

        elif res.text == self.http.body_asserted:
            return True

        # Check if the response body matches a de-asserted state
        if self.http.body_get_deasserted:
            if re.fullmatch(self.http.body_get_deasserted, res.text) is not None:
                return False

        elif res.text == self.http.body_deasserted:
            return False

        raise ExecutionError(f'response does not match asserted or deasserted state: "{res.text}"')



"""
This module contains helper functions for working with version.
"""


def labgrid_version():
    try:
        from .._version import __version__

        return __version__
    except ModuleNotFoundError:
        pass

    import contextlib
    from importlib.metadata import PackageNotFoundError, version

    lg_version = "unknown"

    with contextlib.suppress(PackageNotFoundError):
        lg_version = version("labgrid")

    return lg_version



import sys
import socket
import threading
from time import monotonic, sleep

import gi

gi.require_version("NM", "1.0")
from gi.repository import GLib, NM

# ensure all wrapper objects for Settings types are created
for _name in dir(NM):
    if _name.startswith("Setting"):
        getattr(NM, _name)


class Future:
    def __init__(self):
        self._event = threading.Event()
        self._result = None

    def set(self, result):
        assert self._result is None
        self._result = result
        self._event.set()

    def wait(self):
        self._event.wait()
        if isinstance(self._result, Exception):
            raise self._result
        return self._result


class BackgroundLoop(threading.Thread):
    def __init__(self):
        super().__init__(daemon=True)

    def run(self):
        try:
            GLib.MainLoop(None).run()
        except Exception:
            import traceback

            traceback.print_exc(file=sys.stderr)
            sys.exit(1)

    def block_on(self, func, *args, **kwargs):
        done = threading.Event()
        result = [None]

        def cb(data):
            result[0] = func(*args, **kwargs)
            done.set()

        GLib.main_context_default().invoke_full(GLib.PRIORITY_DEFAULT, cb, None)
        done.wait()
        return result[0]


def address_from_str(s):
    assert "/" in s, "IP address must be in the form address/prefix"
    (addr, prefix) = s.split("/", 1)
    return NM.IPAddress.new(socket.AF_INET6 if ":" in addr else socket.AF_INET, addr, int(prefix))


def connection_from_dict(data):
    con = NM.SimpleConnection.new()
    for setting_name, setting_data in data.items():
        typ = NM.Setting.lookup_type(setting_name).pytype
        setting = typ()
        for k, v in setting_data.items():
            if not v:
                continue

            if k == "mac-address":
                v = ":".join(f"{x:02X}" for x in v)
            elif k == "ssid":
                if isinstance(v, str):
                    v = v.encode()
                v = GLib.Bytes(v)
            elif k == "addresses":
                # setting addresses via the property doesn't seem to work
                for x in v:
                    setting.add_address(address_from_str(x))
                v = None

            if not v:
                continue
            try:
                setting.set_property(k, v)
            except TypeError as e:
                raise TypeError(f"Failed to set propery {k}") from e
        con.add_setting(setting)
    return con


class NMDev:
    def __init__(self, interface):
        self._interface = interface
        self._nm_dev = nm.get_device_by_iface(interface)  # pylint: disable=possibly-used-before-assignment
        if self._nm_dev is None:
            raise ValueError(f"No device found for interface {interface}")

    def _delete_connection(self, con):
        future = Future()

        def cb(con, res, error):
            assert error is None
            try:
                res = con.delete_finish(res)
                future.set(res)
            except Exception as e:
                future.set(e)

        con.delete_async(
            None,  # cancellable
            cb,
            None,
        )

    def get_settings(self):
        lg_con = nm.get_connection_by_id(f"labgrid-{self._interface}")  # pylint: disable=possibly-used-before-assignment
        if lg_con:
            return dict(lg_con.to_dbus(NM.ConnectionSerializationFlags.ALL))

    def get_active_settings(self):
        active_con = self._nm_dev.get_active_connection()
        if active_con:
            con = active_con.get_connection()
            return dict(con.to_dbus(NM.ConnectionSerializationFlags.ALL))

    def configure(self, data):
        lg_con = nm.get_connection_by_id(f"labgrid-{self._interface}")  # pylint: disable=possibly-used-before-assignment
        if lg_con:
            self._delete_connection(lg_con)
        data["connection"].update(
            {
                "id": f"labgrid-{self._interface}",
                "autoconnect": False,
                "interface-name": self._interface,
            }
        )
        con = connection_from_dict(data)

        future = Future()

        def cb(dev, res, error):
            assert error is None
            try:
                res = nm.add_and_activate_connection_finish(res)  # pylint: disable=possibly-used-before-assignment
                future.set(res)
            except Exception as e:
                future.set(e)

        nm.add_and_activate_connection_async(  # pylint: disable=possibly-used-before-assignment
            con,
            self._nm_dev,
            None,  # specific_object
            None,  # cancellable
            cb,
            None,
        )

        future.wait()  # we must wait, but don't need to return it here

    def wait_state(self, expected, timeout):
        res, out_value, _ = NM.utils_enum_from_str(NM.DeviceState, expected)
        if not res:
            raise ValueError(f"invalid state '{expected}'")
        expected = NM.DeviceState(out_value)

        timeout = monotonic() + timeout
        while monotonic() < timeout:
            sleep(0.25)
            if self._nm_dev.get_state() == expected:
                break
        else:
            raise TimeoutError(
                f"state is '{self._nm_dev.get_state().value_nick}' instead of '{expected.value_nick}'"  # pylint: disable=line-too-long
            )

    def disable(self):
        lg_con = nm.get_connection_by_id(f"labgrid-{self._interface}")  # pylint: disable=possibly-used-before-assignment
        if lg_con:
            self._delete_connection(lg_con)

    def _flags_to_str(self, flags):
        return NM.utils_enum_to_str(type(flags), flags)

    def _accesspoint_to_dict(self, ap):
        res = {}

        res["flags"] = self._flags_to_str(ap.get_flags())
        res["wpa-flags"] = self._flags_to_str(ap.get_wpa_flags())
        res["rsn-flags"] = self._flags_to_str(ap.get_rsn_flags())
        res["bssid"] = ap.get_bssid()
        if ap.get_ssid():
            res["ssid"] = ap.get_ssid().get_data().decode(errors="surrogateescape")
        res["frequency"] = ap.get_frequency()
        res["mode"] = self._flags_to_str(ap.get_mode())
        res["max-bitrate"] = ap.get_max_bitrate()
        res["strength"] = ap.get_strength()

        return res

    def get_state(self):
        state = {}

        device = {}
        device["capabilities"] = self._flags_to_str(self._nm_dev.get_capabilities())
        device["device-type"] = self._flags_to_str(self._nm_dev.get_device_type())
        device["state"] = self._flags_to_str(self._nm_dev.get_state())
        device["state-reason"] = self._flags_to_str(self._nm_dev.get_state_reason())
        device["hw-address"] = self._nm_dev.get_hw_address()
        device["driver"] = self._nm_dev.get_driver()
        device["udi"] = self._nm_dev.get_udi()
        device["mtu"] = self._nm_dev.get_mtu()
        device["vendor"] = self._nm_dev.get_vendor()
        device["product"] = self._nm_dev.get_product()
        state["device"] = device

        for name, ip_cfg in (
            ("ip4-config", self._nm_dev.get_ip4_config()),
            ("ip6-config", self._nm_dev.get_ip6_config()),
        ):
            if ip_cfg is None:
                continue
            cfg = {}
            cfg["addresses"] = []
            for address in ip_cfg.get_addresses():
                cfg["addresses"].append(f"{address.get_address()}/{address.get_prefix()}")
            cfg["gateway"] = ip_cfg.get_gateway()
            state[name] = cfg

        try:
            ap = self._nm_dev.get_active_access_point()
        except AttributeError:
            ap = None
        if ap:
            state["active-access-point"] = self._accesspoint_to_dict(ap)

        return state

    def request_scan(self):
        future = Future()

        def cb(dev, res, error):
            assert error is None
            try:
                res = dev.request_scan_finish(res)
                future.set(res)
            except Exception as e:
                future.set(e)

        self._nm_dev.request_scan_async(
            None,  # cancellable
            cb,
            None,
        )

        future.wait()  # we must wait, but don't need to return it here

    def get_access_points(self, scan):
        if scan is None:  # automatically scan if needed
            age = NM.utils_get_timestamp_msec() - self._nm_dev.get_last_scan()
            scan = bool(age > 30_000)

        if scan:
            timeout = monotonic() + 60
            last = self._nm_dev.get_last_scan()
            self.request_scan()

            while monotonic() < timeout:
                sleep(0.25)
                if self._nm_dev.get_last_scan() > last:
                    break
            else:
                raise TimeoutError("scan did not complete")

        aps = self._nm_dev.get_access_points()
        return [self._accesspoint_to_dict(ap) for ap in aps]

    def get_dhcpd_leases(self):
        leases = []
        with open(f"/var/lib/NetworkManager/dnsmasq-{self._interface}.leases") as f:
            for line in f:
                line = line.strip().split()
                if line[3] == "*":
                    line[3] = None
                if line[4] == "*":
                    line[4] = None
                leases.append(
                    {
                        "expire": int(line[0]),
                        "mac": line[1],
                        "ip": line[2],
                        "hostname": line[3],
                        "id": line[4],
                    }
                )
        return leases


if getattr(NM.Client, "__gtype__", None):
    # hide this from sphinx autodoc
    bl = BackgroundLoop()
    bl.start()
    nm = bl.block_on(NM.Client.new, None)

_nmdevs = {}


def _get_nmdev(interface):
    if interface not in _nmdevs:
        _nmdevs[interface] = NMDev(interface)
    return _nmdevs[interface]


def handle_configure(interface, settings):
    nmdev = _get_nmdev(interface)
    return nmdev.configure(settings)


def handle_wait_state(interface, expected, timeout=60):
    nmdev = _get_nmdev(interface)
    return nmdev.wait_state(expected, timeout)


def handle_disable(interface):
    nmdev = _get_nmdev(interface)
    return nmdev.disable()


def handle_get_active_settings(interface):
    nmdev = _get_nmdev(interface)
    return nmdev.get_active_settings()


def handle_get_settings(interface):
    nmdev = _get_nmdev(interface)
    return nmdev.get_settings()


def handle_get_state(interface):
    nmdev = _get_nmdev(interface)
    return nmdev.get_state()


def handle_get_dhcpd_leases(interface):
    nmdev = _get_nmdev(interface)
    return nmdev.get_dhcpd_leases()


def handle_request_scan(interface):
    nmdev = _get_nmdev(interface)
    return nmdev.request_scan()


def handle_get_access_points(interface, scan=None):
    nmdev = _get_nmdev(interface)
    return nmdev.get_access_points(scan)


methods = {
    # basic
    "configure": handle_configure,
    "wait_state": handle_wait_state,
    "disable": handle_disable,
    "get_active_settings": handle_get_active_settings,
    "get_settings": handle_get_settings,
    "get_state": handle_get_state,
    # dhcpd
    "get_dhcpd_leases": handle_get_dhcpd_leases,
    # wireless
    "request_scan": handle_request_scan,
    "get_access_points": handle_get_access_points,
}



import time

import attr


@attr.s(eq=False)
class Timeout:
    """Reperents a timeout (as a deadline)"""
    timeout = attr.ib(
        default=120.0, validator=attr.validators.instance_of(float)
    )

    def __attrs_post_init__(self):
        if self.timeout < 0.0:
            raise ValueError("timeout must be positive")
        self._deadline = time.monotonic() + self.timeout

    @property
    def remaining(self):
        result = self._deadline - time.monotonic()
        return result if result > 0.0 else 0.0

    @property
    def expired(self):
        return self._deadline <= time.monotonic()



import random
import string


# Remove RID to avoid markers containing substrings like ERROR, FAIL, WARN, INFO or DEBUG
MARKER_POOL = tuple(c for c in string.ascii_uppercase if c not in 'RID')

def gen_marker():
    return ''.join(random.choice(MARKER_POOL) for i in range(10))
